from gilda.term import Term
from gilda.grounder import Grounder, filter_for_organism
import pytest
from . import appreq


gr = Grounder()


def test_grounder():
    entries = gr.lookup('kras')
    statuses = [e.status for e in entries]
    assert 'curated' in statuses
    for entry in entries:
        if entry.status == 'curated':
            assert entry.id == '6407', entry

    scores = gr.ground('kras')
    assert len(scores) == 1, scores
    assert appreq(scores[0].score, 0.9845), scores
    scores = gr.ground('k-ras')
    assert len(scores) == 1, scores
    assert appreq(scores[0].score, 0.9936), scores
    scores = gr.ground('KRAS')
    assert len(scores) == 1, scores
    assert appreq(scores[0].score, 1.0), scores
    scores = gr.ground('bRaf')
    assert len(scores) == 1, scores
    assert appreq(scores[0].score, 0.9936), scores


def test_ground_best():
    score = gr.ground_best('kras')
    assert score is not None
    assert appreq(score.score, 0.9845), score
    score = gr.ground_best('k-ras')
    assert score is not None
    assert appreq(score.score, 0.9936), score
    score = gr.ground_best('KRAS')
    assert score is not None
    assert appreq(score.score, 1.0), score
    score = gr.ground_best('bRaf')
    assert score is not None
    assert appreq(score.score, 0.9936), score

    # Check that when no grounding
    # is possible, none is returned
    assert gr.ground_best("nope nope nope") is None


def test_grounder_bug():
    # Smoke test to make sure the 'NA' entry in grounding terms doesn't get
    # turned into a None
    gr.ground('Na')


def test_grounder_num_entries():
    entries = gr.lookup('NPM1')
    assert len(entries) == 5, entries
    entries = gr.lookup('H4')
    assert len(entries) == 7, entries


def test_grounder_depluralize():
    # Note that lookup returns all matches with no de-duplication
    # or filtering so we get two identical FPLX entries and a yeast protein
    # entry here.
    entries = gr.lookup('RAFs')
    assert len(entries) == 9, entries
    for entry in entries:
        assert entry.norm_text == 'raf'


def test_disambiguate_adeft():
    matches = gr.ground('IR')
    matches = gr.disambiguate('IR', matches, 'Insulin Receptor (IR)')
    for match in matches:
        assert match.disambiguation is not None
        assert match.disambiguation['type'] == 'adeft'
        assert match.disambiguation['match'] in ('grounded', 'ungrounded')
        assert match.disambiguation['score'] is not None
        if match.term.db == 'HGNC' and match.term.id == '6091':
            assert match.disambiguation['match'] == 'grounded'
            assert match.disambiguation['score'] == 1.0


def test_disambiguate_gilda():
    matches = gr.ground('NDR1')
    matches = gr.disambiguate('NDR1', matches, 'STK38')
    for match in matches:
        assert match.disambiguation['type'] == 'gilda'
        assert match.disambiguation['match'] == 'grounded'
        if match.term.db == 'HGNC' and match.term.id == '17847':
            assert match.disambiguation['score'] > 0.99
        if match.term.db == 'HGNC' and match.term.id == '7679':
            assert match.disambiguation['score'] < 0.01


def test_rank_namespace():
    matches = gr.ground('interferon-gamma')
    assert matches[0].term.db == 'HGNC'
    assert matches[1].term.db == 'EFO'


def test_rank_namespace_custom_order():
    """Test when applying a custom order.
    See also the above test ``test_rank_namespace``."""
    custom_grounder = Grounder(namespace_priority=["EFO", "HGNC"])
    matches = custom_grounder.ground('interferon-gamma')
    assert matches[0].term.db == 'EFO'
    assert matches[1].term.db == 'HGNC'


def test_aa_synonym():
    matches = gr.ground('WN')
    assert '141447' not in {m.term.id for m in matches}

    matches = gr.ground('W-N')
    assert '141447' not in {m.term.id for m in matches}


def test_organism_filter():
    dummy = 'dummy'
    t1 = Term('x', dummy, dummy, dummy, dummy, dummy, dummy, '9606')
    t2 = Term('x', dummy, dummy, dummy, dummy, dummy, dummy, '10090')
    t3 = Term('x', dummy, dummy, dummy, dummy, dummy, dummy, None)
    terms = filter_for_organism([t1, t2, t3],
                                organisms=['9606'])
    assert len(terms) == 2, terms
    assert {t.organism for t in terms} == {None, '9606'}
    terms = filter_for_organism([t1, t2, t3],
                                organisms=['10090'])
    assert len(terms) == 2, terms
    assert {t.organism for t in terms} == {None, '10090'}
    terms = filter_for_organism([t1, t2, t3],
                                organisms=['10090', '9606'])
    assert len(terms) == 2, terms
    assert {t.organism for t in terms} == {None, '10090'}
    terms = filter_for_organism([t1, t2, t3],
                                organisms=['9606', '10090'])
    assert len(terms) == 2, terms
    assert {t.organism for t in terms} == {None, '9606'}


def test_organisms():
    matches = gr.ground('Raf1')
    assert len(matches) == 2, len(matches)
    organisms = {match.term.organism for match in matches}
    assert organisms == {'9606'}, matches

    matches = gr.ground('Raf1', organisms=['10090'])
    assert len(matches) == 1, len(matches)
    organisms = {match.term.organism for match in matches}
    assert organisms == {'10090'}

    matches = gr.ground('Raf1', organisms=['9606', '10090'])
    assert len(matches) == 2, len(matches)
    organisms = {match.term.organism for match in matches}
    assert organisms == {'9606'}, matches

    matches = gr.ground('Raf1', organisms=['10090', '9606'])
    assert len(matches) == 1, len(matches)
    organisms = {match.term.organism for match in matches}
    assert organisms == {'10090'}, matches


def test_nonhuman_gene_synonyms():
    matches = gr.ground('Tau', organisms=['10090'])
    assert matches[0].term.db == 'UP', matches
    assert matches[0].term.id == 'P10637', matches


def test_uniprot_gene_synonym():
    matches = gr.ground('MEKK2')
    assert matches[0].term.db == 'HGNC', matches
    assert matches[0].term.entry_name == 'MAP3K2'


def test_greek_to_spelled_out():
    matches = gr.ground('interferon-γ')
    assert matches
    assert matches[0].term.entry_name == 'IFNG'


def test_roman_arabic_ground():
    for term in ['neurexin II', 'neurexin 2', 'neurexin-2', 'neurexin-ii']:
        matches = gr.ground(term)
        assert any((m.term.db, m.term.id) == ('HGNC', '8009') for m in matches)


def test_ground_go_activity():
    matches = gr.ground('EGFR')
    assert 'GO' not in {m.term.db for m in matches}, matches


def test_unidecode():
    txt = 'Löfgren’s syndrome'
    matches = gr.ground(txt)
    assert matches[0].term.db, matches[0].term.id == ('EFO', '0009466')

    txts = ['Aymé-Gripp syndrome', 'Ayme-Gripp syndrome']
    for txt in txts:
        matches = gr.ground(txt)
        assert len(matches) == 2
        assert {m.term.db for m in matches} == {'EFO', 'MESH'}

    txts = ['Bi₇O₉I₃', 'Bi7O9I3']
    for txt in txts:
        matches = gr.ground(txt)
        assert len(matches) == 1
        assert (matches[0].term.db, matches[0].term.id) == \
            ('MESH', 'C000605741')


def test_subsumed_terms():
    txt = 'mitochondria'
    matches = gr.ground(txt)
    assert len(matches) == 1
    match = matches[0]
    assert match.term.db == 'GO'
    assert len(match.subsumed_terms) == 1
    assert match.subsumed_terms[0].db == 'GO', match.subsumed_terms[0]
    assert match.subsumed_terms[0].source_db == 'MESH', match.subsumed_terms[0]

    assert match.get_groundings() == {(match.term.db, match.term.id),
                                      (match.subsumed_terms[0].source_db,
                                       match.subsumed_terms[0].source_id)}
    assert match.get_namespaces() == {match.term.db,
                                      match.subsumed_terms[0].source_db}


def test_hgnc_alias_names():
    matches = gr.ground("FP prostanoid receptor")
    assert ('HGNC', '9600') in {(m.term.db, m.term.id) for m in matches}


def test_namespaces():
    matches = gr.ground('KRAS', namespaces=['CHEBI'])
    assert not matches
    matches = gr.ground('KRAS', namespaces=['HGNC'])
    assert matches


def test_sqlite():
    from gilda.resources.sqlite_adapter import build
    matches = gr.ground('braf')
    entries = {'braf': gr.entries['braf']}
    build(entries, 'test.db')

    grsql = Grounder('test.db')
    matchessql = grsql.ground('braf')
    assert len(matches) == len(matchessql)

    assert grsql.ground('kras') == []


def test_strip_whitespace():
    matches = gr.ground(' inflammatory response ')
    assert matches


def test_instantiate():
    """Test instantiating the grounder with different data structures."""
    term = Term(
        "mitochondria",
        "Mitochondria",
        "GO",
        "GO:0005739",
        "mitochondrion",
        "synonym",
        "mesh",
        None,
        "MESH",
        "D008928",
    )

    # test instantiating with list
    gr = Grounder([term])
    assert len(gr.ground("mitochondria")) == 1

    # test instantiating with set
    gr = Grounder({term})
    assert len(gr.ground("mitochondria")) == 1

    # test instantiating with tuple
    gr = Grounder((term,))
    assert len(gr.ground("mitochondria")) == 1

    # test instantiating with iterable
    gr = Grounder(iter([term]))
    assert len(gr.ground("mitochondria")) == 1

    # test instantiating with dict
    gr = Grounder({term.norm_text: [term]})
    assert len(gr.ground("mitochondria")) == 1

    with pytest.raises(TypeError):
        Grounder(5)



__all__ = [
    'ground',
    'get_models',
    'get_names',
    'get_grounder',
    'make_grounder',
    'annotate',
]

from typing import List, Mapping, Union, Optional

from gilda.grounder import Grounder, Annotation
from gilda.term import Term


class GrounderInstance(object):
    def __init__(self):
        self.grounder = None

    def get_grounder(self):
        if self.grounder is None:
            self.grounder = Grounder()
        return self.grounder

    def ground(self, text, context=None, organisms=None,
               namespaces=None):
        return self.get_grounder().ground(text, context=context,
                                          organisms=organisms,
                                          namespaces=namespaces)

    def get_models(self):
        return self.get_grounder().get_models()

    def get_names(self, db, id, status=None, source=None):
        return self.get_grounder().get_names(db, id,
                                             status=status,
                                             source=source)

    @property
    def prefix_index(self):
        return self.get_grounder().prefix_index


grounder = GrounderInstance()


def ground(text, context=None, organisms=None, namespaces=None):
    """Return a list of scored matches for a text to ground.

    Parameters
    ----------
    text : str
        The entity text to be grounded.
    context : Optional[str]
        Any additional text that serves as context for disambiguating the
        given entity text, used if a model exists for disambiguating the
        given text.
    organisms : Optional[List[str]]
        A list of taxonomy identifiers to use as a priority list
        when surfacing matches for proteins/genes from multiple organisms.
    namespaces : Optional[List[str]]
        A list of namespaces to restrict the matches to. By default, no
        restriction is applied.

    Returns
    -------
    list[gilda.grounder.ScoredMatch]
        A list of ScoredMatch objects representing the groundings.

    Examples
    --------
    Ground a string corresponding to an entity name, label, or synonym

    >>> import gilda
    >>> scored_matches = gilda.ground('mapt')

    The matches are sorted in descending order by score, and in the event of
    a tie, by the namespace of the primary grounding. Each scored match has a
    :class:`gilda.term.Term` object that contain information about the primary
    grounding.

    >>> scored_matches[0].term.db
    'hgnc'
    >>> scored_matches[0].term.id
    '6893'
    >>> scored_matches[0].term.get_curie()
    'hgnc:6893'

    The score for each match can be accessed directly:

    >>> scored_matches[0].score
    0.7623

    The rationale for each match is contained in the ``match`` attribute
    whose fields are described in :class:`gilda.scorer.Match`:

    >>> match_object = scored_matches[0].match

    Give optional context to be used by Gilda's disambiguation models, if available

    >>> scored_matches = gilda.ground('ER', context='Calcium is released from the ER.')

    Only return results from a certain namespace, such as when a family and gene have the same name

    >>> scored_matches = gilda.ground('ESR', namespaces=["hgnc"])
    """
    return grounder.ground(text=text, context=context, organisms=organisms, namespaces=namespaces)


def annotate(
    text: str,
    sent_split_fun=None,
    organisms=None,
    namespaces=None,
    context_text: str = None,
) -> List[Annotation]:
    """Annotate a given text with Gilda (i.e., do named entity recognition).

    Parameters
    ----------
    text : str
        The text to be annotated.
    sent_split_fun : Callable[str, Iterable[Tuple[int, int]]], optional
        A function that splits the text into sentences. The default is
        :func:`nltk.tokenize.PunktSentenceTokenizer.span_tokenize`. The function
        should take a string as input and return an iterable of coordinate pairs
        corresponding to the start and end coordinates for each sentence in the
        input text.
    organisms : list[str], optional
        A list of organism names to pass to the grounder. If not provided,
        human is used.
    namespaces : list[str], optional
        A list of namespaces to pass to the grounder to restrict the matches
        to. By default, no restriction is applied.
    context_text :
        A longer span of text that serves as additional context for the text
        being annotated for disambiguation purposes.

    Returns
    -------
    list[Annotation]
        A list of matches where each match is an Annotation object
        which contains as attributes the text span that was matched,
        the list of ScoredMatches, and the start and end character offsets of
        the text span.
    """
    import gilda.ner

    return gilda.ner.annotate(
        text,
        grounder=grounder,
        sent_split_fun=sent_split_fun,
        organisms=organisms,
        namespaces=namespaces,
        context_text=context_text,
    )


def get_models():
    """Return a list of entity texts for which disambiguation models exist.

    Returns
    -------
    list[str]
        The list of entity texts for which a disambiguation model is
        available.
    """
    return grounder.get_models()


def get_names(db, id, status=None, source=None):
    """Return a list of entity texts corresponding to a given database ID.

    Parameters
    ----------
    db : str
        The database in which the ID is an entry, e.g., HGNC.
    id : str
        The ID of an entry in the database.
    status : Optional[str]
        If given, only entity texts with the given status e.g., "synonym"
        are returned.
    source : Optional[str]
        If given, only entity texts from the given source e.g., "uniprot"
        are returned.
    """
    return grounder.get_names(db, id, status=status, source=source)


def get_grounder() -> Grounder:
    """Initialize and return the default Grounder instance.

    Returns
    -------
    :
        A Grounder instance whose attributes and methods can be used
        directly.
    """
    return grounder.get_grounder()


def make_grounder(
    terms: Union[str, List[Term], Mapping[str, List[Term]]],
) -> Grounder:
    """Create a custom grounder from a list of Terms.

    Parameters
    ----------
    terms :
        Specifies the grounding terms that should be loaded in the Grounder.
        If str, it is interpreted as a path to a grounding
        terms gzipped TSV file which is then loaded. If list, it is assumed to
        be a flat list of Terms. If dict, it is assumed to be a grounding terms
        dict with normalized entity strings as keys and lists of Term objects
        as values.
        Default: None

    Returns
    -------
    :
        A Grounder instance, initialized with either the default terms
        loaded from the resource file or a custom set of terms
        if the terms argument was specified.

    Examples
    --------
    The following example shows how to get an ontology with :mod:`obonet` and
    load custom terms:

    .. code-block:: python

        from gilda import make_grounder
        from gilda.process import normalize
        from gilda import Term

        prefix = "UBERON"
        url = "http://purl.obolibrary.org/obo/uberon/basic.obo"
        g = obonet.read_obo(url)
        custom_obo_terms = []
        it = tqdm(g.nodes(data=True), unit_scale=True, unit="node")
        for node, data in it:
            # Skip entries imported from other ontologies
            if not node.startswith(f"{prefix}:"):
                continue

            identifier = node.removeprefix(f"{prefix}:")

            name = data["name"]
            custom_obo_terms.append(gilda.Term(
                norm_text=normalize(name),
                text=name,
                db=prefix,
                id=identifier,
                entry_name=name,
                status="name",
                source=prefix,
            ))

            # Add terms for all synonyms
            for synonym_raw in data.get("synonym", []):
                try:
                    # Try to parse out of the quoted OBO Field
                    synonym = synonym_raw.split('"')[1].strip()
                except IndexError:
                    continue  # the synonym was malformed

                custom_obo_terms.append(gilda.Term(
                    norm_text=normalize(synonym),
                    text=synonym,
                    db=prefix,
                    id=identifier,
                    entry_name=name,
                    status="synonym",
                    source=prefix,
                ))

        custom_grounder = gilda.make_grounder(custom_obo_terms)
        scored_matches = custom_grounder.ground("head")

    Additional examples for loading custom content from OBO Graph JSON,
    :mod:`pyobo`, and more can be found in the `Jupyter notebooks
    <https://github.com/gyorilab/gilda/tree/master/notebooks>`_
    in the Gilda repository on GitHub.
    """
    return Grounder(terms=terms)



import csv
import gzip
import itertools
import logging
from typing import Iterable, Optional, Set, Tuple

__all__ = [
    "Term",
    "get_curie",
    "get_identifiers_curie",
    "get_bioregistry_url",
    "get_identifiers_url",
    "filter_out_duplicates",
    "dump_terms",
]

logger = logging.getLogger(__name__)


class Term(object):
    """Represents a text entry corresponding to a grounded term.

    Attributes
    ----------
    norm_text : str
        The normalized text corresponding to the text entry, used for lookups.
    text : str
        The text entry itself.
    db : str
        The database / name space corresponding to the grounded term.
    id : str
        The identifier of the grounded term within the database / name space.
    entry_name : str
        The standardized name corresponding to the grounded term.
    status : str
        The relationship of the text entry to the grounded term, e.g., synonym.
    source : str
        The source from which the term was obtained.
    organism : Optional[str]
        When the term represents a protein, this attribute provides the
        taxonomy code of the species for the protein.
        For non-proteins, not provided. Default: None
    source_db : Optional[str]
        If the term's db/id was mapped from a different, original db/id
        from a given source, this attribute provides the original db value
        before mapping.
    source_id : Optional[str]
        If the term's db/id was mapped from a different, original db/id
        from a given source, this attribute provides the original ID value
        before mapping.
    """

    def __init__(self, norm_text, text, db, id, entry_name, status, source,
                 organism=None, source_db=None, source_id=None):
        if not text:
            raise ValueError('Text for Term cannot be empty')
        if not norm_text.strip():
            raise ValueError('Normalized text for Term cannot be empty')
        self.norm_text = norm_text
        self.text = text
        self.db = db
        self.id = str(id)
        self.entry_name = entry_name
        self.status = status
        self.source = source
        self.organism = organism
        self.source_db = source_db
        self.source_id = source_id

    def __str__(self):
        return 'Term(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)' % (
            self.norm_text, self.text, self.db, self.id, self.entry_name,
            self.status, self.source, self.organism, self.source_db,
            self.source_id)

    def __repr__(self):
        return str(self)

    def to_json(self):
        """Return the term serialized into a JSON dict."""
        js = {
            'norm_text': self.norm_text,
            'text': self.text,
            'db': self.db,
            'id': self.id,
            'entry_name': self.entry_name,
            'status': self.status,
            'source': self.source,
        }
        if self.organism:
            js['organism'] = self.organism
        if self.source_db:
            js['source_db'] = self.source_db
        if self.source_id:
            js['source_id'] = self.source_id
        return js

    def to_list(self):
        """Return the term serialized into a list of strings."""
        return [self.norm_text, self.text, self.db, self.id,
                self.entry_name, self.status, self.source,
                self.organism, self.source_db, self.source_id]

    def get_curie(self, style='bioregistry') -> str:
        """Return the compact URI for this term.

        Parameters
        ----------
        style : str, optional
            The style of CURIE to return. One of 'bioregistry' (default) or
            'identifiers'. The 'bioregistry' style corresponds to
            bioregistry.io CURIEs, aqnd the 'identifiers' style corresponds to
            identifiers.org CURIEs.

        Returns
        -------
        :
            A normalized CURIE string for this term, or None if it cannot
            be normalized.
        """
        return get_curie(self.db, self.id, style=style)

    def get_identifiers_url(self):
        """Get the full identifiers.org URL for this term."""
        return get_identifiers_url(self.db, self.id)

    def get_bioregistry_url(self):
        """Return a URL for this term that the Bioregistry can resolve.

        Returns
        -------
        :
            A Bioregistry URL string for this term, or None if it cannot
            be created.
        """
        return get_bioregistry_url(self.db, self.id)

    # Backwards compatibility for the misspelled method name
    def get_idenfiers_url(self):  # pragma: no cover - deprecated spelling
        """Return a URL for this term that Identifiers.org can resolve.

        Returns
        -------
        :
            An Identifiers.org URL string for this term, or None if it cannot
            be created.
        """
        return self.get_identifiers_url()

    def get_groundings(self) -> Set[Tuple[str, str]]:
        """Return all groundings for this term, including from a mapped source.

        Returns
        -------
        :
            A set of tuples representing the main grounding for this term,
            as well as any source grounding from which the main grounding
            was mapped.
        """
        groundings = {(self.db, self.id)}
        if self.source_db:
            groundings.add((self.source_db, self.source_id))
        return groundings

    def get_namespaces(self) -> Set[str]:
        """Return all namespaces for this term, including from a mapped source.

        Returns
        -------
        :
            A set of strings including the main namespace for this term,
            as well as any source namespace from which the main grounding
            was mapped.
        """
        namespaces = {self.db}
        if self.source_db:
            namespaces.add(self.source_db)
        return namespaces


def get_curie(db, id, style='bioregistry') -> Optional[str]:
    """Return a normalized CURIE for the given database and identifier.

    The default Gilda configuration uses INDRA's style of databases and
    identifiers. This function is a simple way to normalize these into
    CURIEs that follow the native style of bioregistry.io or
    identifiers.org.

    Parameters
    ----------
    db : str
        The database / namespace of the identifier assuming the default
        Gilda configuration.
    id : str
        The identifier, assuming the default Gilda configuration.
    style : str, optional
        The style of CURIE to return. One of 'bioregistry' (default)
        or 'identifiers'. The 'bioregistry' style corresponds to
        bioregistry.io CURIEs, aqnd the 'identifiers' style corresponds to
        identifiers.org CURIEs.

    Returns
    -------
    :
        A normalized CURIE string, or None if it cannot be normalized.
    """
    curie_pattern = '{db}:{id}'
    if db == 'UP':
        db = 'uniprot'
    id_parts = id.split(':')
    if len(id_parts) == 1:
        return curie_pattern.format(db=db.lower(), id=id)
    elif len(id_parts) == 2:
        # This is for the namespace-embedded-in-LUI case which
        # for OBO ontologies in identifiers.org is uppercased,
        # otherwise we default to lowercase consistent with Bioregistry
        db_norm = id_parts[0].upper() if style == 'identifiers' \
            else id_parts[0].lower()
        return curie_pattern.format(db=db_norm, id=id_parts[-1])


def get_identifiers_url(db, id) -> Optional[str]:
    """Return a URL for this term that Identifiers.org can resolve.

    Parameters
    ----------
    db : str
        The database / namespace of the identifier assuming the default
        Gilda configuration.
    id : str
        The identifier, assuming the default Gilda configuration.

    Returns
    -------
    :
        An Identifiers.org URL string for this term, or None if it cannot
        be created.
    """
    curie = get_curie(db, id, style='identifiers')
    if curie is not None:
        return f'https://identifiers.org/{curie}'


def get_bioregistry_url(db, id) -> Optional[str]:
    """Return a URL that the Bioregistry can resolve.

    Parameters
    ----------
    db : str
        The database / namespace of the identifier assuming the default
        Gilda configuration.
    id : str
        The identifier, assuming the default Gilda configuration.

    Returns
    -------
    :
        A Bioregistry URL string, or None if it cannot be created.
    """
    curie = get_curie(db, id, style='bioregistry')
    if curie is not None:
        return f'https://bioregistry.io/{curie}' if curie else None


def get_identifiers_curie(db, id) -> Optional[str]:
    """Get the full identifiers.org curie for a term."""
    return get_curie(db, id, style='identifiers')


def get_url(db, id):
    """Get the URL for a term based on its curie parts."""
    curie = get_curie(db, id)


def _term_key(term: Term) -> Tuple[str, str, str, str, str]:
    # We include source_id and source_db to avoid losing
    # potentially important links back to mapped source IDs
    # but we have to make sure these are strings since otherwise
    # they could be None which can't be sorted against strings
    source_db = term.source_db or ''
    source_id = term.source_id or ''
    return term.db, term.id, source_db, source_id, term.text


statuses = {'curated': 1, 'name': 2, 'synonym': 3, 'former_name': 4}


def _priority_key(term: Term) -> Tuple[int, int]:
    """
    Prioritize terms (that are pre-grouped by db/id/text) first
    based on status, and if the status is the same, give priority
    to the ones that are from primary resources
    """
    return (
        statuses[term.status],
        0 if term.db.casefold() == term.source.casefold() else 1
    )


def filter_out_duplicates(terms):
    logger.info('Filtering %d terms for uniqueness...' % len(terms))
    new_terms = []
    for _, terms in itertools.groupby(sorted(terms, key=_term_key),
                                      key=_term_key):
        terms = sorted(terms, key=_priority_key)
        new_terms.append(terms[0])
    # Re-sort the terms
    new_terms = sorted(new_terms, key=lambda x: (x.text, x.db, x.id))
    logger.info('Got %d unique terms...' % len(new_terms))
    return new_terms


TERMS_HEADER = ['norm_text', 'text', 'db', 'id', 'entry_name', 'status',
                'source', 'organism', 'source_db', 'source_id']


def dump_terms(terms: Iterable[Term], fname) -> None:
    """Dump a list of terms to a tsv.gz file."""
    logger.info('Dumping into %s', fname)
    with gzip.open(fname, 'wt', encoding='utf-8') as fh:
        writer = csv.writer(fh, delimiter='\t')
        writer.writerow(TERMS_HEADER)
        writer.writerows(t.to_list() for t in terms)



"""Utilities for Pandas."""

from functools import partial
from typing import Optional, Union, TYPE_CHECKING

from .grounder import Grounder
from . import api

if TYPE_CHECKING:
    import pandas

__all__ = [
    "ground_df",
    "ground_df_map",
]


def ground_df(
    df: "pandas.DataFrame",
    source_column: Union[str, int],
    *,
    target_column: Union[None, str, int] = None,
    grounder: Optional[Grounder] = None,
    **kwargs,
) -> None:
    """
    Ground the elements of a column in a Pandas dataframe as CURIEs, in-place.

    Parameters
    ----------
    df :
        A pandas dataframe
    source_column :
        The column to ground. This column contains text corresponding
        to named entities' labels or synonyms
    target_column :
        The column where to put the groundings (either a CURIE string,
        or None). It's possible to create a new column when passing
        a string for this argument. If not given, will create a new
        column name like ``<source column>_grounded``.
    grounder :
        A custom grounder. If none given, uses the built-in grounder.
    kwargs :
        Keyword arguments passed to :meth:`Grounder.ground`, could
        include context, organisms, or namespaces.

    Examples
    --------
    The following example shows how to use this function.

    .. code-block:: python

        import pandas as pd
        import gilda

        url = "https://raw.githubusercontent.com/OBOAcademy/obook/master/docs/tutorial/linking_data/data.csv"
        df = pd.read_csv(url)
        gilda.ground_df(df, source_column="disease", target_column="disease_curie")
    """
    if target_column is None:
        target_column = f"{source_column}_grounded"
    df[target_column] = ground_df_map(
        df=df, source_column=source_column, grounder=grounder, **kwargs,
    )


def ground_df_map(
    df: "pandas.DataFrame",
    source_column: Union[str, int],
    *,
    grounder: Optional[Grounder] = None,
    **kwargs,
) -> "pandas.Series":
    """
    Ground the elements of a column in a Pandas dataframe as CURIEs.

    Parameters
    ----------
    df :
        A pandas dataframe
    source_column :
        The column to ground. This column contains text corresponding
        to named entities' labels or synonyms
    grounder :
        A custom grounder. If none given, uses the built-in ground.
    kwargs :
        Keyword arguments passed to :meth:`Grounder.ground`, could
        include context, organisms, or namespaces.

    Returns
    -------
    series :
        A pandas series representing the grounded CURIE strings.
        Contains NaNs if grounding was not successful or if
        there was an NaN in the cell before.
    """
    if grounder is None:
        grounder = api.grounder
    func = partial(_ground_helper, grounder=grounder, **kwargs)
    series = df[source_column].map(func)
    return series


def _ground_helper(text, grounder: Grounder, **kwargs) -> Optional[str]:
    if not isinstance(text, str):
        return None
    scored_matches = grounder.ground(text, **kwargs)
    if not scored_matches:
        return None
    return scored_matches[0].term.get_curie()



import re
import csv
import obonet
from collections import defaultdict
from gilda.term import Term
from gilda.grounder import Grounder, normalize
from biomappings.resources import append_prediction_tuples

# This is from the Reach bioresources file as a convenient
# source of pre-processed synonyms
UNIPROT = '/Users/ben/src/bioresources/kb/uniprot-proteins.tsv'
PROONTO = '/Users/ben/src/bioresources/kb/protein-ontology-fragments.tsv'
PROONTO_OBO = '/Users/ben/src/bioresources/scripts/pro_reasoned.obo'


def organism_filter(organism):
    if organism == 'Human':
        return True
    if organism == 'SARS-CoV-2':
        return True
    return False


def get_uniprot_terms():
    terms = {}
    with open(UNIPROT, 'r') as fh:
        reader = csv.reader(fh, delimiter='\t')
        for row in reader:
            if '#' in row[1] and organism_filter(row[2]):
                synonym = row[0]
                id = row[1].split('#')[1]
                term = Term(normalize(synonym), synonym, 'UP',
                            id, synonym, 'synonym', 'uniprot')
                terms[str(term.to_json())] = term
    return list(terms.values())


def pre_process_synonym(synonym):
    synonyms = []
    synonyms.append(synonym)
    remove_suffix = re.sub(r'(.*) (\(.*)\)$', '\\1', synonym)
    synonyms.append(remove_suffix)
    return synonyms


def ground_proonto_terms(grounder):
    matches_per_id = defaultdict(list)
    with open(PROONTO, 'r') as fh:
        reader = csv.reader(fh, delimiter='\t')
        for row in reader:
            synonym, id = row
            for text in pre_process_synonym(synonym):
                matches = grounder.ground(text)
                if matches:
                    matches_per_id[id] += matches
    return matches_per_id


def dump_predictions():
    # source prefix, source identifier, source name, relation
    # target prefix, target identifier, target name, type, source
    source_prefix = 'pr'
    target_prefix = 'uniprot.chain'
    relation = 'skos:exactMatch'
    source = 'https://github.com/indralab/gilda/blob/master/scripts/' \
        'generate_uniprot_chain_proonto_mappings.py'
    match_type = 'lexical'
    rows = []
    pro = obonet.read_obo(PROONTO_OBO)
    for pro_id, matches in matches_per_id.items():
        target_id = matches[0].term.id
        target_name = matches[0].term.entry_name
        source_name = pro.nodes[pro_id]['name']
        row = (source_prefix, pro_id, source_name, relation,
               target_prefix, target_id, target_name, match_type,
               0.8, source)
        rows.append(row)
    append_prediction_tuples(rows, deduplicate=True)


if __name__ == '__main__':
    # 1. Parse all the UniProt synonyms that are for human
    # protein fragments into Gilda Terms
    terms = get_uniprot_terms()
    # 2. Instantiate a grounder with these terms
    terms_dict = defaultdict(list)
    for term in terms:
        terms_dict[term.norm_text].append(term)
    grounder = Grounder(terms_dict)
    # 3. Parse all the Protein Ontology synonyms and ground each of them, then
    # store the results
    matches_per_id = ground_proonto_terms(grounder)
    # 4. Dump spreadsheet with non-ambiguous equivalences in BioMappings format
    dump_predictions()


"""This script benchmarks Gilda on the BioCreative VII BioID corpus.
It dumps multiple result tables in the results folder."""
import json
import os
import pathlib
from collections import defaultdict
from copy import deepcopy
from datetime import datetime
from functools import lru_cache
from textwrap import dedent
from typing import Any, Collection, Dict, Iterable, List, Optional, Set, Tuple

import click
import pandas as pd
import pystow
import tabulate
from lxml import etree
from tqdm import tqdm

import famplex
from gilda import __version__
from gilda.grounder import Grounder, logger
from gilda.resources import mesh_to_taxonomy, popular_organisms
from indra.databases.chebi_client import get_chebi_id_from_pubchem
from indra.databases.hgnc_client import get_hgnc_from_entrez
from indra.databases.uniprot_client import get_hgnc_id
from indra.literature import pmc_client, pubmed_client
from indra.ontology.bio import bio_ontology

logger.setLevel('WARNING')

HERE = os.path.dirname(os.path.abspath(__file__))
TAXONOMY_CACHE_PATH = os.path.join(HERE, 'taxonomy_cache.json')
MODULE = pystow.module('gilda', 'biocreative')
URL = 'https://biocreative.bioinformatics.udel.edu/media/store/files/2017/BioIDtraining_2.tar.gz'

tqdm.pandas()

#: A set of the prefix->prefix mappings missing from the bio-ontology
BO_MISSING_XREFS = set()



class BioIDBenchmarker:
    """Used for evaluating gilda using data from BioCreative VI BioID track

    Parameters
    ----------
    grounder :
        Grounder object to use in evaluation. If None, instantiates a grounder
        with default arguments. Default: None
    equivalences :
        Dictionary of mappings between namespaces. Maps strings of the form
        f'{namespace}:{id}' to strings for equivalent groundings. This is
        used to map groundings from namespaces used the the BioID track
        (e.g. Uberon, Cell Ontology, Cellosaurus, NCBI Taxonomy) that are not
        available by default in Gilda. Default: None
    """

    def __init__(
        self,
        *,
        grounder: Optional[Grounder] = None,
        equivalences: Optional[Dict[str, Any]] = None,
    ):
        print("Instantiating benchmarker...")
        if grounder is None:
            grounder = Grounder()
        print("Instantiating bio ontology...")
        bio_ontology.initialize()
        if equivalences is None:
            equivalences = {}
        available_namespaces = set()
        for terms in grounder.entries.values():
            for term in terms:
                available_namespaces.add(term.db)
        self.grounder = grounder
        self.equivalences = equivalences
        self.available_namespaces = list(available_namespaces)
        self.paper_level_grounding = defaultdict(set)
        self.processed_data = self._process_annotations_table()
        if os.path.exists(TAXONOMY_CACHE_PATH):
            with open(TAXONOMY_CACHE_PATH, 'r') as fh:
                self.taxonomy_cache = json.load(fh)
        else:
            self.taxonomy_cache = {}
        print('Taxonomy cache length: %s' % len(self.taxonomy_cache))

    def get_mappings_tables(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Get table showing how goldstandard groundings are being mapped

        Namespaces used in the Bioc dataset may only partially overlap with
        those used by Gilda. Users may pass in a dictionary of equivalences
        mapping groundings used in the Bioc dataset to Gilda's namespaces.
        This method generated tables showing how groundings used in the
        dataset project onto Gilda's namespaces through these equivalences.

        Returns
        -------
        mapping_table : py:class`pandas.DataFrame`
        Rows correspond to namespaces used in the Bioc dataset, columns
        to namespaces used in Gilda (automatically populated based on a
        Gilda Grounders entries attribute). There is also a row Total
        containing the sum of values for all other rows. There are columns
        Count, and Total Mapped, showing the total count of entries for
        each row namespace, and the total count of entries that could be
        mapped to a Gilda namespace respectively.

        The same row namespace can be mapped to multiple column namespaces,
        causing values in them Total Mapped column to be less than the sum of
        values of other columns in the same row. Additionally, in some cases
        an entry in the Bioc dataset has multiple curated groundings, causing
        the counts not to add up to the number of entries in the dataset.

        mapping_table_unique : py:class`pandas.DataFrame`
        Similar to mapping table, but counts are given for unique named
        entity groundings, ignoring duplication of groundings between rows
        in the Bioc dataset.
        """
        # Build dataframes for storing information. Values will be filled in
        # by looping through rows of the dataset.
        index = [get_display_name(ns) for ns in bioc_nmspaces] + ['Total']
        columns = (['Count'] +
                   [get_display_name(ns) for ns in self.available_namespaces] +
                   ['Total Mapped'])
        mapping_table = pd.DataFrame(index=index, columns=columns)
        mapping_table.fillna(0, inplace=True)
        mapping_table_unique = pd.DataFrame(index=index, columns=columns)
        mapping_table_unique.fillna(0, inplace=True)

        # Maps row namespaces to sets of associated grounding ids
        nmspace_ids = defaultdict(set)
        # Maps row namespaces to to set of Gilda grounding ids that have
        # been mapped to from them
        mapped_to_nmspace_ids = defaultdict(set)
        # Maps row namespaces to sets of associated grounding ids, but
        # only in cases where some mapping exists to a Gilda grounding
        mapped_from_nmspace_ids = defaultdict(set)
        # Looping through dataframe is costly. There may be a way to write
        # this with a clever series of groupbys
        for _, row in self.processed_data.iterrows():
            # For each row loop through goldstandard groundings. There can
            # be more than one
            for g1 in row.obj:
                # Get the namespace. If it is not one of the namespaces used
                # in evaluation, discard and continue to the next iteration
                # of the loop
                nmspace1 = g1.split(':', maxsplit=1)[0]
                if nmspace1 not in bioc_nmspaces:
                    continue
                # Increment total count for this namespace
                mapping_table.loc[get_display_name(nmspace1), 'Count'] += 1
                # If this particular grounding has not been seen before for
                # this namespace increment unique count and mark grounding
                # as having been seen
                if g1 not in nmspace_ids[nmspace1]:
                    mapping_table_unique.loc[get_display_name(nmspace1),
                                             'Count'] += 1
                    nmspace_ids[nmspace1].add(g1)
                # Get all of the synonyms that grounding can be mapped to.
                # This includes the grounding itself. If a row namespace is
                # also a column namespace, we consider this to be a valid
                # mapping
                synonyms = self.get_synonym_set([g1])
                # Track which namespaces have been used so we don't overcount
                # when the same grounding can be mapped to multiple groundings
                # in the same namespace
                used_namespaces = set()
                for g2 in synonyms:
                    nmspace2 = g2.split(':', maxsplit=1)[0]
                    # If a namespace mapped to is not available in Gilda
                    # or if we have already tallied a mapping to this namespace
                    # for this particular row, discard and continue
                    if nmspace2 not in self.available_namespaces or \
                        nmspace2 in used_namespaces:
                        continue
                    # If Gilda namespace has not been mapped to in the curent
                    # row increment the count of entries in the namespace with
                    # a mapping to a Gilda namespace
                    if not used_namespaces:
                        mapping_table.loc[get_display_name(nmspace1),
                                          'Total Mapped'] += 1
                    used_namespaces.add(nmspace2)
                    # If the grounding g1 has never been mapped to a Gilda
                    # namespace increment the unique count
                    if g1 not in mapped_from_nmspace_ids[nmspace1]:
                        mapping_table_unique. \
                            loc[get_display_name(nmspace1),
                                'Total Mapped'] += 1
                        mapped_from_nmspace_ids[nmspace1].add(g1)
                    # Increment count for mapping of row namespace to
                    # column namespace
                    mapping_table.loc[get_display_name(nmspace1),
                                      get_display_name(nmspace2)] += 1
                    # If the grounding in column namespace has not been mapped
                    # to by the grounding in row namespace, increment unique
                    # count
                    if g2 not in mapped_to_nmspace_ids[nmspace1]:
                        mapping_table_unique. \
                            loc[get_display_name(nmspace1),
                                get_display_name(nmspace2)] += 1
                        mapped_to_nmspace_ids[nmspace1].add(g2)
        # Generate total rows
        mapping_table.loc['Total', :] = mapping_table.sum()
        mapping_table_unique.loc['Total', :] = mapping_table_unique.sum()
        mapping_table.reset_index(inplace=True)
        mapping_table.rename({'index': 'Namespace'}, inplace=True)
        mapping_table_unique.reset_index(inplace=True)
        mapping_table_unique.rename({'index': 'Namespace'}, inplace=True)
        return mapping_table, mapping_table_unique

    def _process_annotations_table(self):
        """Extract relevant information from annotations table."""
        print("Extracting information from annotations table...")
        df = MODULE.ensure_tar_df(
            url=URL,
            inner_path='BioIDtraining_2/annotations.csv',
            read_csv_kwargs=dict(sep=',', low_memory=False),
        )
        # Split entries with multiple groundings then normalize ids
        df.loc[:, 'obj'] = df['obj'].apply(self._normalize_ids)
        # Add synonyms of gold standard groundings to help match more things
        df.loc[:, 'obj_synonyms'] = df['obj'].apply(self.get_synonym_set)
        # Create column for entity type
        df.loc[:, 'entity_type'] = df.apply(self._get_entity_type_helper, axis=1)
        processed_data = df[['text', 'obj', 'obj_synonyms', 'entity_type',
                             'don_article']]
        print("%d rows in processed annotations table." % len(processed_data))
        processed_data = processed_data[processed_data.entity_type
                                        != 'unknown']
        print("%d rows in annotations table without unknowns." %
              len(processed_data))
        for don_article, text, synonyms in df[['don_article', 'text',
                                               'obj_synonyms']].values:
            self.paper_level_grounding[don_article, text].update(synonyms)
        return processed_data

    def _get_entity_type_helper(self, row) -> str:
        if self._get_entity_type(row.obj) != 'Gene':
            return self._get_entity_type(row.obj)
        elif any(y.startswith('HGNC') for y in row.obj_synonyms):
            return 'Human Gene'
        else:
            return 'Nonhuman Gene'

    def ground_entities_with_gilda(self, context=True, species=True):
        """Compute gilda groundings of entity texts in corpus

        Adds two columns to the internal dataframe for groundings with
        and without context based disambiguation.
        """
        df = self.processed_data
        tqdm.write("Grounding no-context corpus with Gilda...")
        df.loc[:, 'groundings_no_context'] = df.text. \
            progress_apply(self._get_grounding_list)

        if context and species:
            tqdm.write("Grounding with-context corpus with Gilda...")
            # use from tqdm.contrib.concurrent import thread_map
            df.loc[:, 'groundings'] = df. \
                progress_apply(self._get_row_grounding_list_with_model, axis=1)
        elif not context and species:
            df.loc[:, 'groundings'] = df. \
                progress_apply(self._get_row_grounding_list_sans_model, axis=1)
        elif context and not species:
            raise NotImplementedError
        else:
            raise ValueError("why would we ever do this")
            # tqdm.write("Skipping grounding with context.")
            # df.loc[:, 'groundings'] = df.groundings_no_context
        tqdm.write("Finished grounding corpus with Gilda...")
        self._evaluate_gilda_performance()

    def _get_row_grounding_list_with_model(self, row):
        return self._get_grounding_list(
            row.text,
            context=self._get_plaintext(row.don_article),
            organisms=self._get_organism_priority(row.don_article),
        )

    def _get_row_grounding_list_sans_model(self, row):
        return self._get_grounding_list(
            row.text,
            organisms=self._get_organism_priority(row.don_article),
        )

    @lru_cache(maxsize=None)
    def _get_plaintext(self, don_article: str) -> str:
        """Get plaintext content from XML file in BioID corpus

        Parameters
        ----------
        don_article :
            Identifier for paper used within corpus.

        Returns
        -------
        :
            Plaintext of specified article
        """
        directory = MODULE.ensure_untar(url=URL, directory='BioIDtraining_2')
        path = directory.joinpath('BioIDtraining_2', 'fulltext_bioc',
                                  f'{don_article}.xml')
        tree = etree.parse(path.as_posix())
        paragraphs = tree.xpath('//text')
        paragraphs = [' '.join(text.itertext()) for text in paragraphs]
        return '/n'.join(paragraphs) + '/n'

    def _get_organism_priority(self, don_article):
        don_article = str(don_article)
        if don_article in self.taxonomy_cache:
            return self.taxonomy_cache[don_article]
        pubmed_id = pubmed_from_pmc(don_article)
        taxonomy_ids = get_taxonomy_for_pmid(pubmed_id)
        organisms = [o for o in popular_organisms
                     if o in taxonomy_ids] + \
                    [o for o in popular_organisms
                     if o not in taxonomy_ids]
        self.taxonomy_cache[don_article] = organisms
        return organisms

    @classmethod
    def _normalize_ids(cls, curies: str) -> List[str]:
        return [cls._normalize_id(y) for y in curies.split('|')]

    @staticmethod
    def _normalize_id(curie):
        """Convert ID into standardized format, f'{namespace}:{id}'."""
        if curie.startswith('CVCL'):
            return curie.replace('_', ':')
        split_id = curie.split(':', maxsplit=1)
        if split_id[0] == 'Uberon':
            return split_id[1]
        if split_id[0] == 'Uniprot':
            return f'UP:{split_id[1]}'
        if split_id[0] in ['GO', 'CHEBI']:
            return f'{split_id[0]}:{split_id[0]}:{split_id[1]}'
        return curie

    @staticmethod
    def _get_entity_type(groundings: Collection[str]) -> str:
        """Get entity type based on entity groundings of text in corpus."""
        if any(
            grounding.startswith('NCBI gene') or grounding.startswith('UP')
            for grounding in groundings
        ):
            return 'Gene'
        elif any(grounding.startswith('Rfam') for grounding in groundings):
            return 'miRNA'
        elif any(grounding.startswith('CHEBI') or grounding.startswith('PubChem')
                 for grounding in groundings):
            return 'Small Molecule'
        elif any(grounding.startswith('GO') for grounding in groundings):
            return 'Cellular Component'
        elif any(
            grounding.startswith('CVCL') or grounding.startswith('CL')
            for grounding in groundings
        ):
            return 'Cell types/Cell lines'
        elif any(grounding.startswith('UBERON') for grounding in groundings):
            return 'Tissue/Organ'
        elif any(grounding.startswith('NCBI taxon') for grounding in groundings):
            return 'Taxon'
        else:
            return 'unknown'

    def _get_grounding_list(
        self,
        text: str,
        context=None,
        organisms=None,
    ) -> List[Tuple[str, float]]:
        """Run gilda on a text and extract list of result-score tuples."""
        groundings = self.grounder.ground(text, context=context,
                                          organisms=organisms)
        result = []
        for grounding in groundings:
            db, id_ = grounding.term.db, grounding.term.id
            result.append((f'{db}:{id_}', grounding.score))
        return result

    def get_synonym_set(self, curies: Iterable[str]) -> Set[str]:
        """Return set containing all elements in input list along with synonyms
        """
        output = set()
        for curie in curies:
            output.update(self._get_equivalent_entities(curie))
        # We accept all FamPlex terms that cover some or all of the specific
        # entries in the annotations
        covered_fplx = {fplx_entry for fplx_entry, members
                        in fplx_members.items() if (members <= output)}
        output |= {'FPLX:%s' % fplx_entry for fplx_entry in covered_fplx}
        return output

    def _get_equivalent_entities(self, curie: str) -> Set[str]:
        """Return set of equivalent entity groundings

        Uses set of equivalences in self.equiv_map as well as those
        available in indra's hgnc, uniprot, and chebi clients.
        """
        output = {curie}
        prefix, identifier = curie.split(':', maxsplit=1)
        for xref_prefix, xref_id in bio_ontology.get_mappings(prefix, identifier):
            output.add(f'{xref_prefix}:{xref_id}')

        # TODO these should all be in bioontology, eventually
        for xref_curie in self.equivalences.get(curie, []):
            if xref_curie in output:
                continue
            xref_prefix, xref_id = xref_curie.split(':', maxsplit=1)
            if (prefix, xref_prefix) not in BO_MISSING_XREFS:
                BO_MISSING_XREFS.add((prefix, xref_prefix))
                tqdm.write(f'Bioontology v{bio_ontology.version} is missing mappings from {prefix} to {xref_prefix}')
            output.add(xref_curie)

        if prefix == 'NCBI gene':
            hgnc_id = get_hgnc_from_entrez(identifier)
            if hgnc_id is not None:
                output.add(f'HGNC:{hgnc_id}')
        if prefix == 'UP':
            hgnc_id = get_hgnc_id(identifier)
            if hgnc_id is not None:
                output.add(f'HGNC:{hgnc_id}')
        if prefix == 'PubChem':
            chebi_id = get_chebi_id_from_pubchem(identifier)
            if chebi_id is not None:
                output.add(f'CHEBI:{chebi_id}')
        return output

    @staticmethod
    def famplex_isa(hgnc_id: str, fplx_id: str) -> bool:
        """Check if hgnc entity satisfies and isa relation with famplex entity

        Parameters
        ----------
        hgnc_id :
            String of the form f'{namespace}:{id}'
        fplx_id :
            String of the form f'{namespace}:{id}'

        Returns
        -------
        :
            True if hgnc_id corresponds to a valid HGNC grounding,
            fplx_id corresponds to a valid Famplex grounding and the
            former isa the later.
        """
        # TODO can this be swapped directly for the bioontology?
        return famplex.isa('HGNC', hgnc_id, 'FPLX', fplx_id)

    def isa(self, curie_1: str, curie_2: str) -> bool:
        """True if id1 satisfies isa relationship with id2."""
        # if curie_1.startswith('MESH') and curie_2.startswith('MESH'):
        #     return mesh_isa(curie_1, curie_2)
        # # Handle GOGO problem
        # elif curie_1.startswith('GO') and curie_2.startswith('GO'):
        #     curie_1 = curie_1.split(':', maxsplit=1)[1]
        #     curie_2 = curie_2.split(':', maxsplit=1)[1]
        #     try:
        #         return nx.has_path(self.godag, curie_1, curie_2)
        #     except Exception:
        #         return False
        # else:
        #     return curie_1 in self.isa_relations and curie_2 in self.isa_relations[curie_1]
        ns1, id1 = curie_1.split(':', maxsplit=1)
        ns2, id2 = curie_2.split(':', maxsplit=1)
        # TODO did we need to keep some processing on the IDs?
        return bio_ontology.isa(ns1, id1, ns2, id2)

    def top_correct(self, row, disamb=True) -> bool:
        groundings = row.groundings if disamb \
            else row.groundings_no_context
        if not groundings:
            return False
        groundings = [g[0] for g in groundings]
        top_grounding = groundings[0]
        ref_groundings = \
            self.paper_level_grounding[(row.don_article, row.text)]
        return top_grounding in ref_groundings

    def exists_correct(self, row, disamb: bool = True) -> bool:
        groundings = row.groundings if disamb \
            else row.groundings_no_context
        if not groundings:
            return False
        groundings = {g[0] for g in groundings}
        ref_groundings = \
            self.paper_level_grounding[(row.don_article, row.text)]
        return len(groundings & ref_groundings) > 0

    def top_correct_w_fplx(self, row, disamb: bool = True) -> bool:
        groundings = row.groundings if disamb \
            else row.groundings_no_context
        if not groundings:
            return False
        groundings = [g[0] for g in groundings]
        top_grounding = groundings[0]
        ref_groundings = \
            self.paper_level_grounding[(row.don_article, row.text)]
        return any(
            x == top_grounding or self.famplex_isa(x, top_grounding)
            for x in ref_groundings
        )

    def exists_correct_w_fplx(self, row, disamb: bool = True) -> bool:
        groundings = row.groundings if disamb \
            else row.groundings_no_context
        if not groundings:
            return False
        groundings = [g[0] for g in groundings]
        ref_groundings = \
            self.paper_level_grounding[(row.don_article, row.text)]
        return any(
            x == y or self.famplex_isa(x, y)
            for x in ref_groundings
            for y in groundings
        )

    def top_correct_loose(self, row, disamb=True) -> bool:
        groundings = row.groundings if disamb \
            else row.groundings_no_context
        if not groundings:
            return False
        groundings = [g[0] for g in groundings]
        top_grounding = groundings[0]
        ref_groundings = \
            self.paper_level_grounding[(row.don_article, row.text)]
        return any(
            x == top_grounding
            or self.isa(x, top_grounding)
            or self.isa(top_grounding, x)
            for x in ref_groundings
        )

    def exists_correct_loose(self, row, disamb=True) -> bool:
        groundings = row.groundings if disamb \
            else row.groundings_no_context
        if not groundings:
            return False
        groundings = [g[0] for g in groundings]
        ref_groundings = \
            self.paper_level_grounding[(row.don_article, row.text)]
        return any(
            x == y
            or self.isa(x, y)
            or self.isa(y, x)
            for x in ref_groundings
            for y in groundings
        )

    def _evaluate_gilda_performance(self):
        """Calculate statistics showing Gilda's performance on corpus

        Directly updates internal dataframe
        """
        print("Evaluating performance...")
        df = self.processed_data
        df.loc[:, 'top_correct'] = df.apply(self.top_correct, axis=1)
        df.loc[:, 'top_correct_w_fplx'] = df.apply(self.top_correct_w_fplx, axis=1)
        df.loc[:, 'top_correct_loose'] = df.apply(self.top_correct_loose, axis=1)
        df.loc[:, 'exists_correct'] = df.apply(self.exists_correct, axis=1)
        df.loc[:, 'exists_correct_w_fplx'] = df. \
            apply(self.exists_correct_w_fplx, axis=1)
        df.loc[:, 'exists_correct_loose'] = df. \
            apply(self.exists_correct_loose, axis=1)
        df.loc[:, 'top_correct_no_context'] = df. \
            apply(lambda row: self.top_correct(row, False), axis=1)
        df.loc[:, 'top_correct_w_fplx_no_context'] = df. \
            apply(lambda row: self.top_correct_w_fplx(row, False), axis=1)
        df.loc[:, 'top_correct_loose_no_context'] = df. \
            apply(lambda row: self.top_correct_loose(row, False), axis=1)
        df.loc[:, 'Has Grounding'] = df.groundings. \
            apply(lambda x: len(x) > 0)
        print("Finished evaluating performance...")

    def get_results_tables(
        self,
        match: Optional[str] = 'loose',
        with_context: bool = True,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Get tables of results

        Parameters
        ----------
        match :
            One of 'strict', 'w_fplex', or 'loose'. 'strict' only counts
            a Gilda grounding as a match if it is an exact match or equivalent
            to at least one of the curated groundings for the entry
            (some entries have multiple equivalent curated groundings).
            'w_fplex' also counts a Gilda grounding as a match if the curated
            grounding has an HGNC equivalent and satisfies and isa relationship
            with the Gilda grounding. 'loose' counts the pair x, y of a curated
            grounding and a Gilda groundings as a match if x isa y or y isa x
            within the FPLX, MESH, and GO ontologies, or within the dictionary
            of isa_relations.

        Returns
        -------
        counts_table : py:class:`pandas.DataFrame`
        precision_recall : py:class:`pandas.DataFrame`
        disamb_table : py:class:`pandas.DataFrame`
        """
        if match not in ['strict', 'w_fplex', 'loose']:
            raise ValueError("match must be one of 'strict', 'w_famplex', or"
                             " 'loose'.")
        df = self.processed_data
        if 'top_correct' not in df.columns:
            raise RuntimeError('Gilda groundings have not been computed')
        res_df = df[['entity_type', 'top_correct', 'top_correct_no_context',
                     'exists_correct', 'top_correct_w_fplx',
                     'top_correct_w_fplx_no_context', 'exists_correct_w_fplx',
                     'top_correct_loose', 'top_correct_loose_no_context',
                     'exists_correct_loose', 'Has Grounding']].copy()
        res_df.loc[:, 'Total'] = True
        total = res_df.drop('entity_type', axis=1).sum()
        total = total.to_frame().transpose()
        total.loc[:, 'entity_type'] = 'Total'
        stats = res_df.groupby('entity_type', as_index=False).sum()
        stats = stats[stats['entity_type'] != 'unknown']
        stats = stats.append(total, ignore_index=True)
        stats.loc[:, stats.columns[1:]] = stats[stats.columns[1:]].astype(int)
        if match == 'strict':
            score_cols = ['top_correct', 'exists_correct']
        else:
            score_cols = [f'top_correct_{match}', f'exists_correct_{match}']
        if not with_context:
            # OLD
            # score_cols[0] = score_cols[0] + ['_no_context']
            score_cols[0] = score_cols[0] + '_no_context'
        cols = ['entity_type'] + score_cols + ['Has Grounding', 'Total']
        counts_table = deepcopy(stats[cols])
        new_column_names = ['Entity Type', 'Correct', 'Exists Correct',
                            'Has Grounding', 'Total']
        counts_table.columns = new_column_names
        precision_recall = pd.DataFrame(
            index=stats.index,
            columns=[
                'Entity Type',
                'Precision',
                'Exists Correct PR',
                'Recall',
                'Exists Correct RC',
            ],
        )

        precision_recall.loc[:, 'Entity Type'] = counts_table['Entity Type']
        precision_recall.loc[:, 'Precision'] = \
            round(counts_table['Correct'] /
                  counts_table['Has Grounding'], 3)
        precision_recall.loc[:, 'Exists Correct PR'] = \
            round(counts_table['Exists Correct'] /
                  counts_table['Has Grounding'], 3)
        precision_recall.loc[:, 'Recall'] = round(counts_table['Correct'] /
                                                  counts_table['Total'], 3)
        precision_recall.loc[:, 'Exists Correct RC'] = \
            round(counts_table['Exists Correct'] / counts_table['Total'], 3)

        precision_recall.loc[:, 'Correct F1'] = \
            f1(precision_recall.loc[:, 'Precision'],
               precision_recall.loc[:, 'Recall'])
        precision_recall.loc[:, 'Exists Correct F1'] = \
            f1(precision_recall.loc[:, 'Exists Correct PR'],
               precision_recall.loc[:, 'Exists Correct RC'])

        cols = ['entity_type', 'top_correct_loose_no_context',
                'top_correct_loose', 'exists_correct', 'Total']
        new_column_names = ['Entity Type', 'Correct', 'Correct (disamb)',
                            'Exists Correct', 'Total']
        disamb_table = stats[cols]
        disamb_table.columns = new_column_names
        return counts_table, precision_recall, disamb_table


def get_taxonomy_for_pmid(pmid: str) -> Set[str]:
    if not pmid:
        return set()
    import time
    tqdm.write(f'Looking up annotations for pmid:{pmid}')
    time.sleep(2)
    mesh_annots = pubmed_client.get_mesh_annotations(pmid)
    if mesh_annots is None:
        return set()
    mesh_ids = {annot['mesh'] for annot in mesh_annots}
    taxonomy_ids = set()
    for mesh_id in mesh_ids:
        if mesh_id in mesh_to_taxonomy:
            taxonomy_ids.add(mesh_to_taxonomy[mesh_id])
        mesh_parents = [id for ns, id in
                        bio_ontology.get_parents('MESH', mesh_id)]
        for mesh_parent in mesh_parents:
            if mesh_parent in mesh_to_taxonomy:
                taxonomy_ids.add(mesh_to_taxonomy[mesh_parent])
    tqdm.write('-----')
    tqdm.write('PMID: %s' % pmid)
    for mesh_annot in mesh_annots:
        tqdm.write(mesh_annot['text'])
    tqdm.write('Taxonomy IDs: %s' % taxonomy_ids)
    tqdm.write('-----')
    return taxonomy_ids


def pubmed_from_pmc(pmc_id):
    pmc_id = str(pmc_id)
    if not pmc_id.startswith('PMC'):
        pmc_id = f'PMC{pmc_id}'
    ids = pmc_client.id_lookup(pmc_id, 'pmcid')
    pmid = ids.get('pmid')
    return pmid


#: Namespaces used in Bioc dataset after standardization
bioc_nmspaces = ['UP', 'NCBI gene', 'Rfam', 'CHEBI', 'PubChem', 'GO',
                 'CL', 'CVCL', 'UBERON', 'NCBI taxon']

#: Mapping of namespaces to row and column names. Namespaces not
#: included will be used as row and column names unmodifed.
nmspace_displaynames = {
    'UP': 'Uniprot', 'NCBI gene': 'Entrez',
    'PubChem': 'PubChem', 'CL': 'Cell Ontology',
    'CVCL': 'Cellosaurus', 'UBERON': 'Uberon',
    'FPLX': 'Famplex'
}


def get_famplex_members():
    from indra.databases import hgnc_client
    fplx_entities = famplex.load_entities()
    fplx_children = defaultdict(set)
    for fplx_entity in fplx_entities:
        members = famplex.individual_members('FPLX', fplx_entity)
        for db_ns, db_id in members:
            if db_ns == 'HGNC':
                db_id = hgnc_client.get_current_hgnc_id(db_id)
                if db_id:
                    fplx_children[fplx_entity].add('%s:%s' % (db_ns, db_id))
    return dict(fplx_children)


fplx_members = get_famplex_members()


# NOTE: these mappings are already integrated into equivalences.json
def get_uberon_mesh_mappings():
    import obonet
    from indra.databases import mesh_client
    g = obonet.read_obo('/Users/ben/src/uberon/src/ontology/uberon-edit.obo')
    mappings = {}
    for node, data in g.nodes(data=True):
        xrefs = [x[5:] for x in data.get('xref', []) if x.startswith('MESH')]
        if len(xrefs) != 1:
            continue
        xref = xrefs[0]
        if mesh_client.get_mesh_name(xref, offline=True):
            mappings[node] = 'MESH:%s' % xref
    return mappings


# NOTE: these mappings are already integrated into equivalences.json
def get_cl_mesh_mappings():
    import re
    classdef_prefix = "# Class: obo:"
    mesh_id_pattern = re.compile(r'MESH:[CD][0-9]+')
    mappings = {}
    with open('/Users/ben/src/cell-ontology/src/ontology/cl-edit.owl', 'r') as fh:
        node = None
        for line in fh:
            if line.startswith(classdef_prefix):
                node_owl = line[len(classdef_prefix):len(classdef_prefix) + 10]
                node = node_owl.replace('_', ':')
            mesh_ids = set(mesh_id_pattern.findall(line))
            if node and len(mesh_ids) == 1:
                mappings[node] = list(mesh_ids)[0]
    return mappings


def get_display_name(ns: str) -> str:
    """Gets row/column name associated to a namespace"""
    return nmspace_displaynames[ns] if ns in nmspace_displaynames else ns


def f1(precision: float, recall: float) -> float:
    """Calculate the F1 score."""
    return 2 * precision * recall / (precision + recall)


@click.command()
@click.option(
    '--data',
    type=click.Path(dir_okay=True, file_okay=False),
    default=os.path.join(HERE, 'data'),
)
@click.option(
    '--results',
    type=click.Path(dir_okay=True, file_okay=False),
    default=os.path.join(HERE, 'results', "bioid_performance", __version__),
)
@click.option("--no-model-disambiguation", is_flag=True)
@click.option("--no-species-disambiguation", is_flag=True)
def main(data: str, results: str, no_model_disambiguation: bool, no_species_disambiguation: bool):
    """Run this script to evaluate gilda on the BioCreative VI BioID corpus.

    It has two optional arguments, --datapath and --resultspath that specify
    the path to the directory with necessary data and the path to the
    directory where results will be stored. The directory at datapath must
    contain a folder BIoIDtraining_2 containing the BIoID training corpus,
    and can optionally contain files equivalences.json
    serializing dictionaries of equivalence and isa relations between
    groundings. Results files will be added to the results directory in
    timestamped files.

    The data can be downloaded from
    https://biocreative.bioinformatics.udel.edu/media/store/files/2017/BioIDtraining_2.tar.gz,
    and needs to be extracted in the benchmarks/data folder.
    """
    data_path = os.path.expandvars(os.path.expanduser(data))
    results_path = os.path.expandvars(os.path.expanduser(results))
    os.makedirs(data_path, exist_ok=True)
    os.makedirs(results_path, exist_ok=True)

    try:
        with open(os.path.join(data_path, 'equivalences.json')) as f:
            equivalences = json.load(f)
    except FileNotFoundError:
        equivalences = {}
    benchmarker = BioIDBenchmarker(equivalences=equivalences)
    benchmarker.ground_entities_with_gilda(
        context=not no_model_disambiguation,
        species=not no_species_disambiguation,
    )
    print("Constructing mappings table...")
    mappings_table, mappings_table_unique = benchmarker.get_mappings_tables()
    print("Constructing results table...")
    counts, precision_recall, disamb_table = \
        benchmarker.get_results_tables(match='strict')
    print(
        f"Gilda v{__version__}, Bio-ontology v{bio_ontology.version},"
        f" model-based disambiguation={not no_model_disambiguation},"
        f" species-based disambiguation={not no_species_disambiguation}"
    )
    print(precision_recall.to_markdown(index=False))
    time = datetime.now().strftime('%y%m%d-%H%M%S')
    if no_model_disambiguation and no_species_disambiguation:
        outname = f'benchmark_no_disambiguation_{time}'
    elif no_model_disambiguation and not no_species_disambiguation:
        outname = f'benchmark_no_model_disambiguation_{time}'
    elif not no_model_disambiguation and no_species_disambiguation:
        outname = f'benchmark_no_species_disambiguation_{time}'
    else:
        outname = f'benchmark_{time}'

    # Generate output document
    caption0 = dedent(f"""\
    # Gilda Benchmarking

    Bio-ontology: v{bio_ontology.version}
    Gilda: v{__version__}
    Date: {time}
    """)
    caption1 = dedent("""\
        ## Table 1

        Mapping of groundings for entities in BioID corpus into namespaces used by
        Gilda. Count is by entries in corpus with groundings being counted multiple
        times if they occur in more than one entry. Some entries contain multiple
        equivalent curated groundings, leading to a discrepancy between the counts
        shown here and those in the other tables.
    """)
    table1 = mappings_table.to_markdown(index=False)
    caption2 = dedent("""\
        ## Table 2

        Mapping of groundings for entities in BioID corpus into Namespaces used by
        Gilda. Count is by unique groundings, with the same grounding only being
        counted once even if it appears in many entries.
    """)
    table2 = mappings_table_unique.to_markdown(index=False)
    caption3 = dedent("""\
        ## Table 3

        Counts of number of entries in corpus for each entity type, along with
        number of entries where Gilda's top grounding is correct, the number
        where one of Gilda's groundings is correct, and the number of entries
        where Gilda produced some grounding. Context based disambiguation is
        applied and Gilda's groundings are considered correct if there is
        an isa relation between the gold standard grounding and Gilda's or
        vice versa.
    """)
    table3 = counts.to_markdown(index=False)
    caption4 = dedent("""\
        ## Table 4

        Precision and recall values for Gilda performance by entity type. Values
        are given both for the case where Gilda is considered correct only if the
        top grounding matches and the case where Gilda is considered correct if
        any of its groundings match.
    """)
    table4 = precision_recall.to_markdown(index=False)
    caption5 = dedent("""\
        ## Table 5

        Comparison of results with and without context based disambiguation.
    """)
    table5 = disamb_table.to_markdown(index=False)
    output = '\n\n'.join([
        caption0,
        caption1, table1,
        caption2, table2,
        caption3, table3,
        caption4, table4,
        caption5, table5,
    ])
    result_stub = pathlib.Path(results_path).joinpath(outname)
    md_path = result_stub.with_suffix(".md")
    with open(md_path, 'w') as f:
        f.write(output)
    print(f'Output summary at {md_path}')

    latex_output = dedent(f'''\
    \\section{{Tables}}
        {mappings_table.to_latex(index=False, caption=caption1, label='tab:mappings')}
        {mappings_table_unique.to_latex(index=False, caption=caption2, label='tab:mappings-unique')}
        {counts.to_latex(index=False, caption=caption3, label='tab:counts')}
        {precision_recall.round(3).to_latex(index=False, caption=caption4, label='tab:precision-recall')}
        {disamb_table.to_latex(index=False, caption=caption5, label='tab:disambiguation')}
    ''')
    latex_path = result_stub.with_suffix(".tex")
    with open(latex_path, 'w') as file:
        print(latex_output, file=file)

    tsv_path = result_stub.with_suffix(".tsv")
    json_path = result_stub.with_suffix(".json")
    benchmarker.processed_data.to_csv(tsv_path, sep='\t', index=False)
    benchmarker.processed_data.to_json(json_path, force_ascii=False, orient="records", indent=2)


if __name__ == '__main__':
    main()



from gilda.term import Term
from gilda.generate_terms import parse_uniprot_synonyms, \
    filter_out_duplicates, get_terms_from_uniprot_row


def test_parse_embedded_parentheses_uniprot():
    txt = ('Sodium-coupled neutral amino acid transporter 4 (Amino acid '
           'transporter A3) (Na(+)-coupled neutral amino acid transporter 4)'
           ' (Solute carrier family 38 member 4) (System A amino acid '
           'transporter 3)')
    syms = parse_uniprot_synonyms(txt)
    for s in syms:
        assert not s.endswith(' '), '"%s"' % s
    assert len(syms) == 5, syms
    assert 'System A amino acid transporter 3' in syms


def test_parse_embedded_parentheses_uniprot_2():
    txt = ('Na(+)/H(+) exchange regulatory cofactor NHE-RF1 (NHERF-1) '
           '(Ezrin-radixin-moesin-binding phosphoprotein 50) (EBP50)'
           ' (Regulatory cofactor of Na(+)/H(+) exchanger) '
           '(Sodium-hydrogen exchanger regulatory factor 1) '
           '(Solute carrier family 9 isoform A3 regulatory factor 1)')
    syms = parse_uniprot_synonyms(txt)
    for s in syms:
        assert not s == '+/', '"%s"' % s
    assert len(syms) == 7, syms
    assert 'Na(+)/H(+) exchange regulatory cofactor NHE-RF1' in syms, syms


def test_parse_embedded_parentheses_uniprot_3():
    txt = ('Solute carrier family 13 member 3 (Na(+)/dicarboxylate ' \
           'cotransporter 3, NaDC-3, rNaDC3) (Sodium-dependent high-affinity' \
           ' dicarboxylate transporter 2)')
    syms = parse_uniprot_synonyms(txt)
    assert syms == \
        ['Solute carrier family 13 member 3',
         'Na(+)/dicarboxylate cotransporter 3, NaDC-3, rNaDC3',
         'Sodium-dependent high-affinity dicarboxylate transporter 2']


def test_parse_parentheses_in_name():
    txt = 'DNA (cytosine-5)-methyltransferase 1 (EC:2.1.1.37)'
    syms = parse_uniprot_synonyms(txt)
    assert syms == ['DNA (cytosine-5)-methyltransferase 1', 'EC:2.1.1.37'], \
        syms


def test_parse_parantheses_in_name2():
    txt = ('Neutral amino acid transporter B(0) (ATB(0)) '
           '(Baboon M7 virus receptor) (RD114/simian type D retrovirus '
           'receptor) (Sodium-dependent neutral amino acid transporter type 2) '
           '(Solute carrier family 1 member 5)')
    # The challenge here is the trailing (0) at the end of the first
    # synonym which is part of that synonym and shouldn't be treated as
    # a separate synonym.
    syms = parse_uniprot_synonyms(txt)
    assert syms == \
        ['Neutral amino acid transporter B(0)',
         'ATB(0)',
         'Baboon M7 virus receptor',
         'RD114/simian type D retrovirus receptor',
         'Sodium-dependent neutral amino acid transporter type 2',
         'Solute carrier family 1 member 5']


def test_filter_priority():
    term1 = Term('mekk2', 'MEKK2', 'HGNC', '6854', 'MAP3K2',
                 'former_name', 'hgnc', '9606')
    term2 = Term('mekk2', 'MEKK2', 'HGNC', '6854', 'MAP3K2',
                 'synonym', 'up', '9606')
    terms = filter_out_duplicates([term1, term2])
    assert len(terms) == 1
    term = terms[0]
    assert term.status == 'synonym'


def test_filter_priority_by_source():
    term1 = Term('mekk2', 'MEKK2', 'HGNC', '6854', 'MAP3K2',
                 'synonym', 'hgnc', '9606')
    term2 = Term('mekk2', 'MEKK2', 'HGNC', '6854', 'MAP3K2',
                 'synonym', 'up', '9606')
    terms = filter_out_duplicates([term1, term2])
    assert len(terms) == 1
    assert terms[0] == term1

    # now test the other way, to make sure order doesn't matter
    terms = filter_out_duplicates([term2, term1])
    assert len(terms) == 1
    assert terms[0] == term1


def test_get_terms_simple():
    row = {'Entry': 'P15056',
           'Gene Names (primary)': 'BRAF',
           'Gene Names (synonym)': 'BRAF1 RAFB1',
           'Protein names':
               ('Serine/threonine-protein kinase B-raf '
                '(EC 2.7.11.1) (Proto-oncogene B-Raf) (p94) '
                '(v-Raf murine sarcoma viral oncogene homolog B1)'),
           'Organism (ID)': '9606'}
    terms = get_terms_from_uniprot_row(row)
    assert len(terms) == 7, terms
    assert all(term.db == 'HGNC' for term in terms), terms


def test_get_terms_multi_gene_human():
    row = {'Entry': 'P62805',
           'Gene Names (primary)':
               ('H4C1; H4C2; H4C3; H4C4; H4C5; H4C6; H4C8;'
                ' H4C9; H4C11; H4C12; H4C13; H4C14; H4C15; H4-16'),
           'Gene Names (synonym)': ('H4/A H4FA HIST1H4A; H4/I H4FI HIST1H4B; '
                                      'H4/G H4FG HIST1H4C; H4/B H4FB HIST1H4D; '
                                      'H4/J H4FJ HIST1H4E; H4/C H4FC HIST1H4F; '
                                      'H4/H H4FH HIST1H4H; H4/M H4FM HIST1H4I; '
                                      'H4/E H4FE HIST1H4J; H4/D H4FD HIST1H4K; '
                                      'H4/K H4FK HIST1H4L; H4/N H4F2 H4FN '
                                      'HIST2H4 HIST2H4A; H4/O H4FO HIST2H4B; '
                                      'HIST4H4'),
           'Protein names': 'Histone H4',
           'Organism (ID)': '9606'}
    terms = get_terms_from_uniprot_row(row)
    assert len(terms) == 1, terms
    assert terms[0].db == 'UP'
    assert terms[0].text == 'Histone H4'


def test_get_terms_multi_gene_nonhuman():
    row = {'Entry': 'P62784',
           'Gene Names (primary)': ('his-1; his-5; his-10; his-14; his-18; '
                                      'his-26; his-28; his-31; his-37; his-38; '
                                      'his-46; his-50; his-56; his-60; '
                                      'his-64; his-67'),
           'Gene Names (synonym)': '; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ',
           'Protein names': 'Histone H4',
           'Organism (ID)': '6239'}
    terms = get_terms_from_uniprot_row(row)
    assert len(terms) == 17



import matplotlib.pyplot as plt
from gilda.grounder import load_gilda_models


if __name__ == '__main__':
    models = load_gilda_models(-1)
    f1s = [m.stats['f1']['mean'] for m in models.values()]
    plt.ion()
    plt.figure(figsize=(8, 4))
    plt.hist(f1s, 100, color='gray')
    plt.xlabel('Mean F1 of disambiguation model', fontsize=18)
    plt.ylabel('Number of models', fontsize=18)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.axvline(x=0.7, linestyle='--', color='gray')
    plt.subplots_adjust(left=0.114, right=0.98, bottom=0.157, top=0.97)
    plt.show()
    plt.savefig('gilda_model_f1_plots.pdf')


from gilda.process import get_capitalization_pattern
from gilda.scorer import generate_match, score_string_match
from . import appreq


def test_cap_pattern():
    assert get_capitalization_pattern('Abcd', True) == 'sentence_initial_cap'
    assert get_capitalization_pattern('A') == 'single_cap_letter'
    assert get_capitalization_pattern('BRAF') == 'all_caps'
    assert get_capitalization_pattern('abcd') == 'all_lower'
    assert get_capitalization_pattern('Abcd') == 'initial_cap'
    assert get_capitalization_pattern('aBcd') == 'mixed'


def test_generate_match():
    match = generate_match('BRAF', 'Braf')
    assert ('all_caps', 'initial_cap') in match.cap_combos, match

    match = generate_match('Braf', 'BRAF')
    assert ('initial_cap', 'all_caps') in match.cap_combos, match

    match = generate_match('BrAf', 'braf')
    assert ('mixed', 'all_lower') in match.cap_combos, match


def test_characterize_match_dashes():
    match = generate_match('eIF-5A', 'EIF5A')
    assert 'query' in match.dash_mismatches
    assert ('mixed', 'all_caps') in match.cap_combos


def test_string_match_scoring():
    cases = [
        ('k-ras', 'K-ras', 0.9443),
        ('k-ras', 'KRAS', 0.8589),
        ('k-ras', 'Kras', 0.9424),
        ('ras', 'RAs', 0.6938),
        ('ras', 'ras', 1),
        ('Ras', 'RA-s', 0.9425),
        ('ras', 'RAS', 0.3599)
    ]
    for query, ref, expected_score in cases:
        match = generate_match(query, ref)
        score = score_string_match(match)
        assert appreq(score, expected_score), (score, expected_score)




from gilda.grounder import Grounder
from indra.databases import mesh_client

gr = Grounder()


def get_ambiguities(skip_assertions=True, skip_names=True):
    ambig_entries = {}
    for terms in gr.entries.values():
        for term in terms:
            # We consider it an ambiguity if the same text entry appears
            # multiple times
            key = term.text
            if key in ambig_entries:
                ambig_entries[key].append(term)
            else:
                ambig_entries[key] = [term]
    # It's only an ambiguity if there are two entries at least
    ambig_entries = {k: v for k, v in ambig_entries.items() if len(v) >= 2}

    ambigs = []
    for text, entries in ambig_entries.items():
        dbs = {e.db for e in entries}
        db_ids = {(e.db, e.id) for e in entries}
        statuses = {e.status for e in entries}
        sources = {e.source for e in entries}
        names = {e.entry_name for e in entries}
        # If the entries all point to the same ID, we skip it
        if len(db_ids) <= 1:
            continue
        # If there is a name in statuses, we skip it because it's prioritized
        if skip_names and 'name' in statuses:
            continue
        # We skip curated terms because they are prioritized anyway
        if skip_assertions and 'curated' in statuses:
            continue
        # We can't get CHEBI PMIDs yet
        if 'CHEBI' in dbs:
            continue
        if 'adeft' in sources:
            continue
        # This typically happens for some UniProt IDs that don't have gene
        # names
        if any(n is None for n in names):
            continue
        # Everything else is an ambiguity
        ambigs.append(entries)
    return ambigs


def filter_out_shared_prefix(ambigs):
    non_long_prefix_ambigs = []
    for terms in ambigs:
        names = [term.entry_name for term in terms]
        prefix = lcp(names)
        if len(prefix) < 3:
            non_long_prefix_ambigs.append(terms)
    return non_long_prefix_ambigs


def filter_out_duplicates(ambigs):
    unique_ambigs = []
    for terms in ambigs:
        keys = set()
        unique_terms = []
        for term in terms:
            key = (term.db, term.id)
            if key not in keys:
                keys.add(key)
                unique_terms.append(term)
        unique_ambigs.append(unique_terms)
    return unique_ambigs


def filter_out_mesh_proteins(ambigs):
    mesh_protein = 'D000602'
    mesh_enzyme = 'D045762'
    all_new_ambigs = []
    for terms in ambigs:
        new_ambigs = []
        for term in terms:
            if term.db == 'MESH':
                if mesh_client.mesh_isa(term.id, mesh_protein) or \
                        mesh_client.mesh_isa(term.id, mesh_enzyme):
                    continue
            new_ambigs.append(term)
        if len(new_ambigs) >= 2:
            all_new_ambigs.append(new_ambigs)
        else:
            terms_str = '> ' + '\n> '.join(str(t) for t in terms)
            print('Filtered out:\n%s' % terms_str)
    return all_new_ambigs


def find_families(ambigs):
    # Find possible families
    for ambig in ambigs:
        names = [term.entry_name for term in ambig]
        prefix = lcp(names)
        if len(prefix) < 3:
            continue
        print('%s -> %s' % (ambig[0].text, ', '.join(sorted(names))))


def lcp(strs):
    # This function helps identify hypothetical families based on shared
    # prefixes
    prefix = ''
    for letters in zip(*strs):
        if len(set(letters)) == 1:
            prefix += letters[0]
        else:
            break
    return prefix




import os
import csv
import json
import gzip
import logging
import itertools
import collections.abc
import tempfile
from pathlib import Path
from collections import defaultdict, Counter
from textwrap import dedent
from typing import Iterator, List, Mapping, Optional, Set, Tuple, Union, Iterable
from urllib.request import urlretrieve

from adeft.disambiguate import load_disambiguator
from adeft.modeling.classify import load_model_info
from adeft import available_shortforms as available_adeft_models
from .term import Term, get_curie, get_bioregistry_url
from .process import normalize, replace_dashes, replace_greek_uni, \
    replace_greek_latin, replace_greek_spelled_out, depluralize, \
    replace_roman_arabic
from .scorer import Match, generate_match, score
from .resources import get_gilda_models, get_grounding_terms

__all__ = [
    "Grounder",
    "GrounderInput",
    "ScoredMatch",
    "Annotation",
    "load_terms_file",
    "load_entries_from_terms_file",
    "filter_for_organism",
    "load_adeft_models",
    "load_gilda_models",
]

logger = logging.getLogger(__name__)


GrounderInput = Union[str, Path, Iterable[Term], Mapping[str, List[Term]]]

#: The default namespace priority order
DEFAULT_NAMESPACE_PRIORITY = [
    'FPLX', 'HGNC', 'UP', 'CHEBI', 'GO', 'MESH', 'DOID', 'HP', 'EFO'
]


class Grounder(object):
    """Class to look up and ground query texts in a terms file.

    Parameters
    ----------
    terms :
        Specifies the grounding terms that should be loaded in the Grounder.

        - If ``None``, the default grounding terms are loaded from the
          versioned resource folder.
        - If :class:`str` or :class:`pathlib.Path`, it is interpreted
          as a path to a grounding terms gzipped TSV file which is then
          loaded. If it's a str and looks like a URL, will be downloaded from the internet
        - If :class:`dict`, it is assumed to be a grounding terms dict with
          normalized entity strings as keys and :class:`gilda.term.Term`
          instances as values.
        - If :class:`list`, :class:`set`, :class:`tuple`, or any other iterable,
          it is assumed to be a flat list of
          :class:`gilda.term.Term` instances.
    namespace_priority :
        Specifies a term namespace priority order. For example, if multiple
        terms are matched with the same score, will use this list to decide
        which are given by which namespace appears further towards the front
        of the list. By default, :data:`DEFAULT_NAMESPACE_PRIORITY` is used,
        which, for example, prioritizes famplex entities over HGNC ones.
    """

    entries: Mapping[str, List[Term]]
    namespace_priority: List[str]

    def __init__(
        self,
        terms: Optional[GrounderInput] = None,
        *,
        namespace_priority: Optional[List[str]] = None,
    ):
        if terms is None:
            terms = get_grounding_terms()

        if isinstance(terms, str) and terms.startswith("http"):
            with tempfile.TemporaryDirectory() as directory:
                path = Path(directory).joinpath("terms.tsv.gz")
                urlretrieve(terms, path)  # noqa:S310
                self.entries = load_terms_file(path)
        elif isinstance(terms, (str, Path)):
            extension = os.path.splitext(terms)[1]
            if extension == '.db':
                from .resources.sqlite_adapter import SqliteEntries
                self.entries = SqliteEntries(terms)
            else:
                self.entries = load_terms_file(terms)
        elif isinstance(terms, dict):
            self.entries = terms
        elif isinstance(terms, collections.abc.Iterable):
            self.entries = defaultdict(list)
            for term in terms:
                self.entries[term.norm_text].append(term)
            self.entries = dict(self.entries)
        else:
            raise TypeError('terms is neither a path nor a list of terms,'
                            'nor a normalized entry name to term dictionary')

        self.prefix_index = {}
        self._build_prefix_index()

        self.adeft_disambiguators = find_adeft_models()
        self.gilda_disambiguators = None

        self.namespace_priority = (
            DEFAULT_NAMESPACE_PRIORITY
            if namespace_priority is None else
            namespace_priority
        )

    def _build_prefix_index(self):
        prefix_index = defaultdict(set)
        for norm_term in self.entries:
            if not norm_term:
                continue
            parts = norm_term.split()
            if not parts:
                continue
            prefix_index[parts[0]].add(len(parts))
        self.prefix_index = dict(prefix_index)

    def lookup(self, raw_str: str) -> List[Term]:
        """Return matching Terms for a given raw string.

        Parameters
        ----------
        raw_str :
            A string to be looked up in the set of Terms that the Grounder
            contains.

        Returns
        -------
        :
            A list of Terms that are potential matches for the given string.
        """
        lookups = self._generate_lookups(raw_str)
        entries = []
        for lookup in lookups:
            entries += self.entries.get(lookup, [])
        return entries

    def _generate_lookups(self, raw_str: str) -> Set[str]:
        # TODO: we should propagate flags about depluralization and possible
        #  other modifications made here and take them into account when
        #  scoring
        # We first add the normalized string itself
        norm = normalize(raw_str)
        lookups = {norm}
        # Then we add a version with dashes replaced by spaces
        norm_spacedash = normalize(replace_dashes(raw_str, ' '))
        lookups.add(norm_spacedash)
        # We then try to replace spelled out greek letters with
        # their unicode equivalents or their latin equivalents
        greek_replaced = normalize(replace_greek_uni(raw_str))
        lookups.add(greek_replaced)
        greek_replaced = normalize(replace_greek_latin(raw_str))
        lookups.add(greek_replaced)
        greek_replaced = normalize(replace_greek_spelled_out(raw_str))
        lookups.add(greek_replaced)
        # We try exchanging roman and arabic numerals
        roman_arabic = normalize(replace_roman_arabic(raw_str))
        lookups.add(roman_arabic)
        # Finally, we attempt to depluralize the word
        for singular, rule in depluralize(raw_str):
            lookups.add(normalize(singular))

        logger.debug('Looking up the following strings: %s' %
                     ', '.join(lookups))
        return lookups

    def _score_namespace(self, term) -> int:
        """Apply a priority to the term based on its namespace.

        .. note::

            This is currently not included as an explicit score term.
            It is just used to rank identically scored entries.
        """
        try:
            return len(self.namespace_priority) - self.namespace_priority.index(term.db)
        except ValueError:
            return 0

    def ground_best(
        self,
        raw_str: str,
        context: Optional[str] = None,
        organisms: Optional[List[str]] = None,
        namespaces: Optional[List[str]] = None,
    ) -> Optional["ScoredMatch"]:
        """Return the best scored grounding for a given raw string.

        Parameters
        ----------
        raw_str : str
            A string to be grounded with respect to the set of Terms that the
            Grounder contains.
        context : Optional[str]
            Any additional text that serves as context for disambiguating the
            given entity text, used if a model exists for disambiguating the
            given text.
        organisms : Optional[List[str]]
            An optional list of organism identifiers defining a priority
            ranking among organisms, if genes/proteins from multiple
            organisms match the input. If not provided, the default
            ['9606'] i.e., human is used.
        namespaces : Optional[List[str]]
            A list of namespaces to restrict matches to. This will apply to
            both the primary namespace of a matched term, to any subsumed
            matches, and to the source namespaces of terms if they were
            created using cross-reference mappings. By default, no
            restriction is applied.

        Returns
        -------
        Optional[gilda.grounder.ScoredMatch]
            The best ScoredMatch returned by :meth:`ground` if any are returned,
            otherwise None.
        """
        scored_matches = self.ground(
            raw_str=raw_str,
            context=context,
            organisms=organisms,
            namespaces=namespaces,
        )
        if scored_matches:
            # Because of the way the ground() function is implemented,
            # the first element is guaranteed to have the best score
            # (after filtering by namespace)
            return scored_matches[0]
        return None

    def ground(self, raw_str, context=None, organisms=None,
               namespaces=None):
        """Return scored groundings for a given raw string.

        Parameters
        ----------
        raw_str : str
            A string to be grounded with respect to the set of Terms that the
            Grounder contains.
        context : Optional[str]
            Any additional text that serves as context for disambiguating the
            given entity text, used if a model exists for disambiguating the
            given text.
        organisms : Optional[List[str]]
            An optional list of organism identifiers defining a priority
            ranking among organisms, if genes/proteins from multiple
            organisms match the input. If not provided, the default
            ['9606'] i.e., human is used.
        namespaces : Optional[List[str]]
            A list of namespaces to restrict matches to. This will apply to
            both the primary namespace of a matched term, to any subsumed
            matches, and to the source namespaces of terms if they were
            created using cross-reference mappings. By default, no
            restriction is applied.

        Returns
        -------
        list[gilda.grounder.ScoredMatch]
            A list of ScoredMatch objects representing the groundings sorted
            by decreasing score.
        """
        if not organisms:
            organisms = ['9606']
        # Stripping whitespaces is done up front directly on the raw string
        # so that all lookups and comparisons are done with respect to the
        # stripped string
        raw_str = raw_str.strip()
        # Initial lookup of all possible matches
        entries = self.lookup(raw_str)
        logger.debug('Filtering %d entries by organism' % len(entries))
        entries = filter_for_organism(entries, organisms)
        logger.debug('Comparing %s with %d entries' %
                     (raw_str, len(entries)))
        # For each entry to compare to, we generate a match data structure
        # describing the comparison of the raw (unnormalized) input string
        # and the entity text corresponding to the matched Term. This match
        # is then further scored to account for the nature of the grounding
        # itself.
        scored_matches = []
        for term in entries:
            match = generate_match(raw_str, term.text)
            sc = score(match, term)
            scored_match = ScoredMatch(term, sc, match)
            scored_matches.append(scored_match)

        # Return early if we don't have anything to avoid calling other
        # functions with no matches
        if not scored_matches:
            return scored_matches

        # Merge equivalent matches
        unique_scores = self._merge_equivalent_matches(scored_matches)

        # If there's context available, disambiguate based on that
        if context:
            unique_scores = self.disambiguate(raw_str, unique_scores, context)

        # Then sort by decreasing score
        rank_fun = lambda x: (x.score, self._score_namespace(x.term))
        unique_scores = sorted(unique_scores, key=rank_fun, reverse=True)

        # If we have a namespace constraint, we filter to the given
        # namespaces.
        if namespaces:
            unique_scores = [
                scored_match for scored_match in unique_scores
                if scored_match.get_namespaces() & set(namespaces)
            ]

        return unique_scores

    def disambiguate(self, raw_str, scored_matches, context):
        # This is only called if context was passed in so we do lazy
        # loading here
        if self.gilda_disambiguators is None:
            self.gilda_disambiguators = load_gilda_models()
        # If we don't have a disambiguator for this string, we return with
        # the original scores intact. Otherwise, we attempt to disambiguate.
        if raw_str in self.adeft_disambiguators:
            logger.info('Running Adeft disambiguation for %s' % raw_str)
            try:
                scored_matches = \
                    self.disambiguate_adeft(raw_str, scored_matches, context)
            except Exception as e:
                logger.exception(e)
        elif raw_str in self.gilda_disambiguators:
            logger.info('Running Gilda disambiguation for %s' % raw_str)
            try:
                scored_matches = \
                    self.disambiguate_gilda(raw_str, scored_matches, context)
            except Exception as e:
                logger.exception(e)

        return scored_matches

    def disambiguate_adeft(self, raw_str, scored_matches, context):
        # We find the disambiguator for the given string and pass in
        # context
        if self.adeft_disambiguators[raw_str] is None:
            self.adeft_disambiguators[raw_str] = load_disambiguator(raw_str)
        res = self.adeft_disambiguators[raw_str].disambiguate([context])
        # The actual grounding dict is at this index in the result
        grounding_dict = res[0][2]
        logger.debug('Result from Adeft: %s' % str(grounding_dict))
        # We attempt to get the score for the 'ungrounded' entry
        ungrounded_score = grounding_dict.get('ungrounded', 1.0)
        # Now we check if each scored match has a corresponding Adeft
        # grounding and score. If we find one, we multiply the original
        # match score with the Adeft score. Otherwise, we multiply the
        # original score with the 'ungrounded' score given by Adeft.
        for match in scored_matches:
            has_adeft_grounding = False
            for grounding, score in grounding_dict.items():
                # There is a corner case here where grounding is
                # some name other than 'ungrounded' but is not a proper
                # ns:id pair.
                if grounding == 'ungrounded' or ':' not in grounding:
                    continue
                db, id = grounding.split(':', maxsplit=1)
                if match.term.db == db and match.term.id == id:
                    match.disambiguation = {'type': 'adeft',
                                            'score': score,
                                            'match': 'grounded'}
                    match.multiply(score)
                    has_adeft_grounding = True
                    break
            if not has_adeft_grounding:
                match.disambiguation = {'type': 'adeft',
                                        'score': ungrounded_score,
                                        'match': 'ungrounded'}
                match.multiply(ungrounded_score)
        return scored_matches

    def disambiguate_gilda(self, raw_str, scored_matches, context):
        res = self.gilda_disambiguators[raw_str].predict_proba([context])
        if not res:
            raise ValueError('No result from disambiguation.')
        grounding_dict = res[0]
        for match in scored_matches:
            key = '%s:%s' % (match.term.db, match.term.id)
            score_entry = grounding_dict.get(key, None)
            score = score_entry if score_entry is not None else 0.0
            match.disambiguation = {'type': 'gilda',
                                    'score': score,
                                    'match': ('grounded'
                                              if score_entry is not None
                                              else 'ungrounded')}
            match.multiply(score)
        return scored_matches

    @staticmethod
    def _merge_equivalent_matches(scored_matches):
        unique_entries = []
        # Characterize an entry by its grounding
        term_dbid = lambda x: (x.term.db, x.term.id)
        # Sort and group scores by grounding
        scored_matches.sort(key=term_dbid)
        entry_groups = itertools.groupby(scored_matches, key=term_dbid)
        # Now look at each group and find the highest scoring match
        for _, entry_group in entry_groups:
            entries = sorted(list(entry_group), key=lambda x: x.score,
                             reverse=True)
            entries[0].subsumed_terms = [e.term for e in entries[1:]]
            unique_entries.append(entries[0])
        # Return the list of unique entries
        return unique_entries

    def get_models(self):
        """Return a list of entity texts for which disambiguation models exist.

        Returns
        -------
        list[str]
            The list of entity texts for which a disambiguation model is
            available.
        """
        if self.gilda_disambiguators is None:
            self.gilda_disambiguators = load_gilda_models()
        return sorted(list(self.gilda_disambiguators.keys()))

    def get_names(self, db, id, status=None, source=None):
        """Return a list of entity texts corresponding to a given database ID.

        Parameters
        ----------
        db : str
            The database in which the ID is an entry, e.g., HGNC.
        id : str
            The ID of an entry in the database.
        status : Optional[str]
            If given, only entity texts with the given status e.g., "synonym"
            are returned.
        source : Optional[str]
            If given, only entity texts from the given source e.g., "uniprot"
            are returned.

        Returns
        -------
        names: list[str]
            A list of entity texts corresponding to the given database/ID
        """
        names = set()
        for entries in self.entries.values():
            for entry in entries:
                if (entry.db == db) and (entry.id == id) and \
                   (not status or entry.status == status) and \
                   (not source or entry.source == source):
                    names.add(entry.text)
        return sorted(names)

    def get_ambiguities(self,
                        skip_names: bool = True,
                        skip_curated: bool = True,
                        skip_name_matches: bool = True,
                        skip_species_ambigs: bool = True) -> List[List[Term]]:
        """Return a list of ambiguous term groups in the grounder.

        Parameters
        ----------
        skip_names :
            If True, groups of terms where one has the "name" status are
            skipped. This makes sense usually since these are prioritized over
            synonyms anyway.
        skip_curated :
            If True, groups of terms where one has the "curated" status
            are skipped. This makes sense usually since these are prioritized
            over synonyms anyway.
        skip_name_matches :
            If True, groups of terms that all share the same standard name
            are skipped. This is effective at eliminating spurious ambiguities
            due to unresolved cross-references between equivalent terms
            in different namespaces.
        skip_species_ambigs :
            If True, groups of terms that are all genes or proteins, and are
            all from different species (one term from each species) are skipped.
            This is effective at eliminating ambiguities between orthologous
            genes in different species that are usually resolved using the
            organism priority list.
        """
        ambig_entries = defaultdict(list)
        for terms in self.entries.values():
            for term in terms:
                # We consider it an ambiguity if the same text entry appears
                # multiple times
                key = term.text
                ambig_entries[key].append(term)

        # It's only an ambiguity if there are two entries at least
        ambig_entries = {k: v for k, v in ambig_entries.items()
                         if len(v) >= 2}

        ambigs = []
        for text, entries in ambig_entries.items():
            dbs = {e.db for e in entries}
            db_ids = {(e.db, e.id) for e in entries}
            statuses = {e.status for e in entries}
            sources = {e.source for e in entries}
            names = {e.entry_name for e in entries}
            # If the entries all point to the same ID, we skip it
            if len(db_ids) <= 1:
                continue
            # If there is a name in statuses, we skip it because it's
            # prioritized
            if skip_names and 'name' in statuses:
                continue
            # We skip curated terms because they are prioritized anyway
            if skip_curated and 'curated' in statuses:
                continue
            # If there is an adeft model already, we skip it
            if 'adeft' in sources:
                continue
            if skip_name_matches:
                if len({e.entry_name.lower() for e in entries}) == 1:
                    continue
            if skip_species_ambigs:
                if dbs <= {'HGNC', 'UP'} and \
                        len({e.organism for e in entries}) == len(entries):
                    continue
            # Everything else is an ambiguity
            ambigs.append(entries)
        return ambigs

    def _iter_terms(self):
        for terms in self.entries.values():
            yield from terms

    def summary_str(self) -> str:
        """Summarize the contents of the grounder."""
        namespaces = {ns for term in self._iter_terms() for ns in term.get_namespaces()}
        status_counter = dict(Counter(term.status for term in self._iter_terms()))
        return dedent(f"""\
        Lookups: {len(self.entries):,}
        Terms: {sum(len(terms) for terms in self.entries.values()):,}
        Term Namespaces: {namespaces}
        Term Statuses: {status_counter}
        Adeft Disambiguators: {len(self.adeft_disambiguators):,}
        Gilda Disambiguators: {len(self.gilda_disambiguators):,}
        """)

    def print_summary(self, **kwargs) -> None:
        """Print the summary of this grounder."""
        print(self.summary_str(), **kwargs)


class ScoredMatch(object):
    """Class representing a scored match to a grounding term.

    Attributes
    -----------
    term : gilda.grounder.Term
        The Term that the scored match is for.
    score : float
        The score associated with the match.
    match : gilda.scorer.Match
        The Match object characterizing the match to the Term.
    disambiguation : Optional[dict]
        Meta-information about disambiguation, when available.
    subsumed_terms : Optional[list[gilda.grounder.Term]]
        A list of additional Term objects that also matched, have the same
        db/id value as the term associated with the match, but were further
        down the score ranking. In some cases examining the subsumed terms
        associated with a match can provide additional metadata in
        downstream applications.
    """
    def __init__(self, term: Term, score, match: Match, disambiguation=None,
                 subsumed_terms=None):
        self.term = term
        self.url = term.get_bioregistry_url()
        self.score = score
        self.match = match
        self.disambiguation = disambiguation
        self.subsumed_terms = subsumed_terms if subsumed_terms else None

    def __str__(self):
        disamb_str = '' if self.disambiguation is None else \
            (',disambiguation=' + json.dumps(self.disambiguation))
        return 'ScoredMatch(%s,%s,%s%s)' % \
            (self.term, self.score, self.match, disamb_str)

    def __repr__(self):
        return str(self)

    def to_json(self):
        js = {
            'term': self.term.to_json(),
            'url': self.url,
            'score': self.score,
            'match': self.match.to_json()
        }
        if self.disambiguation is not None:
            js['disambiguation'] = self.disambiguation
        if self.subsumed_terms:
            js['subsumed_terms'] = [term.to_json()
                                    for term in self.subsumed_terms]
        return js

    def multiply(self, value):
        logger.debug('Multiplying the score of "%s" with %.3f'
                     % (self.term.entry_name, value))
        self.score = self.score * value

    def get_namespaces(self) -> Set[str]:
        """Return all namespaces for this match including from mapped and
        subsumed terms.

        Returns
        -------
        :
            A set of strings representing namespaces for terms involved in
            this match, including the namespace for the primary term as well
            as any subsumed terms, and groundings that come from having
            mapped an original source grounding during grounding resource
            construction.
        """
        return {ns for ns, _ in self.get_groundings()}

    def get_groundings(self) -> Set[Tuple[str, str]]:
        """Return all groundings for this match including from mapped and
        subsumed terms.

        Returns
        -------
        :
            A set of tuples representing groundings for this match including
            the grounding for the primary term as well as any subsumed
            terms, and groundings that come from having mapped an original
            source grounding during grounding resource construction.
        """
        term_groundings = self.term.get_groundings()
        if self.subsumed_terms:
            for sub_term in self.subsumed_terms:
                term_groundings |= sub_term.get_groundings()
        return term_groundings

    def get_grounding_dict(self) -> Mapping[str, str]:
        """Get the groundings as CURIEs and URLs."""
        return {
            get_curie(db, db_id, style='bioregistry'):
                get_bioregistry_url(db, db_id)
            for db, db_id in self.get_groundings()
        }


class Annotation:
    """A class representing a named entity annotation in a given text.

    Attributes
    ----------
    text : str
        The text span that was annotated.
    matches : list[ScoredMatch]
        The list of scored matches for the text span.
    start : int
        The start character offset of the text span.
    end : int
        The end character offset of the text span.
    """
    def __init__(self, text: str, matches: List[ScoredMatch], start: int,
                 end: int):
        self.text = text
        self.matches = matches
        self.start = start
        self.end = end

    def __repr__(self):
        return str(self)

    def __str__(self):
        return (f"Annotation({self.text}, {self.matches}, {self.start}, "
                f"{self.end})")

    def to_json(self):
        """Convert the Annotation object to JSON."""
        return {
            'text': self.text,
            'matches': [match.to_json() for match in self.matches],
            'start': self.start,
            'end': self.end
        }


def load_entries_from_terms_file(terms_file: Union[str, Path]) -> Iterator[Term]:
    """Yield Terms from a compressed terms TSV file path.

    Parameters
    ----------
    terms_file :
        Path to a compressed TSV terms file with columns corresponding to the
        serialized elements of a Term.

    Returns
    -------
    :
        Terms loaded from the file yielded by a generator.
    """
    with gzip.open(terms_file, 'rt', encoding='utf-8') as fh:
        entries = {}
        reader = csv.reader(fh, delimiter='\t')
        # Skip header
        next(reader)
        for row in reader:
            row_nones = [r if r else None for r in row]
            yield Term(*row_nones)


def load_terms_file(terms_file: Union[str, Path]) -> Mapping[str, List[Term]]:
    """Load a TSV file containing terms into a lookup dictionary.

    Parameters
    ----------
    terms_file :
        Path to a compressed TSV terms file with columns corresponding to the
        serialized elements of a Term.

    Returns
    -------
    :
        A lookup dictionary whose keys are normalized entity texts, and values
        are lists of Terms with that normalized entity text.
    """
    entries = {}
    for term in load_entries_from_terms_file(terms_file):
        if term.norm_text in entries:
            entries[term.norm_text].append(term)
        else:
            entries[term.norm_text] = [term]
    return entries


def filter_for_organism(terms, organisms):
    # First we organize terms by organism, including None
    terms_by_organism = defaultdict(list)
    for term in terms:
        # We filter out any organisms that aren't in the list provided
        if term.organism is not None and term.organism not in organisms:
            continue
        terms_by_organism[term.organism].append(term)
    # We first take the terms without organism
    all_terms = terms_by_organism[None]
    # We now find the top organism for which we have at least
    # one term and then add the corresponding terms to the list
    # of all terms
    if set(terms_by_organism) != {None}:
        top_organism = min(set(terms_by_organism) - {None},
                           key=lambda x: organisms.index(x))
        all_terms += terms_by_organism[top_organism]
    return all_terms


def find_adeft_models():
    adeft_disambiguators = {}
    for shortform in available_adeft_models:
        adeft_disambiguators[shortform] = None
    return adeft_disambiguators


def load_adeft_models():
    return {shortform: load_disambiguator(shortform)
            for shortform in find_adeft_models()}


def load_gilda_models(cutoff=0.7):
    with gzip.open(get_gilda_models(), 'rt') as fh:
        models = {k: load_model_info(v)
                  for k, v in json.loads(fh.read()).items()
                  if v['stats']['f1']['mean'] > cutoff}
    return models



from textwrap import dedent
import csv
from io import StringIO

from flask import Blueprint, render_template, request
from flask_wtf import FlaskForm
from wtforms import SelectMultipleField, StringField, SubmitField, TextAreaField
from wtforms.validators import DataRequired
import markupsafe

from gilda.app.proxies import grounder
from gilda import __version__ as version
from gilda.resources import organism_labels, popular_organisms

__all__ = [
    "ui_blueprint",
]

ORGANISMS_FIELD = SelectMultipleField(
    "Species priority (optional)",
    choices=[(org, organism_labels[org]) for org in popular_organisms],
    id="organism-select",
    description=dedent(
        """\
        Optionally select one or more taxonomy
        species IDs to define a species priority list.  Click
        <a type="button" href="#" data-toggle="modal" data-target="#species-modal">
        here <i class="far fa-question-circle">
        </i></a> for more details.
    """
    ),
)


class GroundForm(FlaskForm):
    text = StringField(
        "Entity text",
        validators=[DataRequired()],
        description='Input the entity text (e.g., <span class="bg-info">k-ras</span>) to ground.',
    )
    context = TextAreaField(
        "Context (optional)",
        description=dedent(
            """\
            Optionally provide additional text context to help disambiguation. Click
            <a type="button" href="#" data-toggle="modal" data-target="#context-modal">
            here <i class="far fa-question-circle">
            </i></a> for more details.
        """
        ),
    )
    organisms = ORGANISMS_FIELD
    submit = SubmitField("Submit")

    def get_matches(self):
        return grounder.ground(
            self.text.data, context=self.context.data, organisms=self.organisms.data
        )


class NERForm(FlaskForm):
    text = TextAreaField(
        "Text",
        validators=[DataRequired()],
        description=dedent(
            """\
            <p>Enter text to annotate with named entities and corresponding identifiers.
            The results will highlight recognized entities and provide 
            text position, grounding information and normalized name for each 
            entity.</p> <p>Click 
            <a href="#" onclick="fillExample(); return false;">here</a>
            for an example input text from an
            <a href="https://pubmed.ncbi.nlm.nih.gov/21040840/">article</a> 
            about small G-proteins."""
        ),
        render_kw={"rows": 6},
    )
    organisms = ORGANISMS_FIELD
    submit = SubmitField("Submit")

    def get_annotations(self):
        from gilda.ner import annotate

        return annotate(
            self.text.data, grounder=grounder, organisms=self.organisms.data
        )


ui_blueprint = Blueprint("ui", __name__, url_prefix="/")


@ui_blueprint.route("/", methods=["GET", "POST"])
def home():
    text = request.args.get("text")
    if text is not None:
        context = request.args.get("context")
        organisms = request.args.getlist("organisms")
        matches = grounder.ground(text, context=context, organisms=organisms)
        return render_template(
            "matches.html", matches=matches, version=version, text=text, context=context
        )

    form = GroundForm()
    if form.validate_on_submit():
        matches = form.get_matches()
        return render_template(
            "matches.html",
            matches=matches,
            version=version,
            text=form.text.data,
            context=form.context.data,
            # Add a new form that doesn't auto-populate
            form=GroundForm(formdata=None),
        )
    return render_template("home.html", form=form, version=version)


@ui_blueprint.route("/ner", methods=["GET", "POST"])
def view_ner():
    form = NERForm()
    if form.validate_on_submit():
        annotations = form.get_annotations()
        annotated_text = get_annotated_text(form.text.data, annotations)

        # Generate CSV data
        si = StringIO()
        writer = csv.writer(si)
        writer.writerow([
            "Start", "End", "Text", "Grounding", "Standard Name", "Score",
            "Additional Groundings", "URL"
        ])

        # Write data
        for annotation in annotations:
            match = annotation.matches[0]
            match_curie = match.term.get_curie()
            additional_groundings = ", ".join(
                curie for curie in match.get_grounding_dict().keys()
                if curie != match_curie
            )

            writer.writerow([
                f"{annotation.start}",
                f"{annotation.end}",
                annotation.text,
                match_curie,
                match.term.entry_name,
                f"{match.score:.4f}",
                additional_groundings,
                match.url
            ])

        csv_data = si.getvalue()

        return render_template(
            "ner_matches.html",
            annotations=annotations,
            version=version,
            text=annotated_text,
            csv_data=csv_data,
            # Add a new form that doesn't auto-populate
            form=NERForm(formdata=None),
        )
    return render_template("ner_home.html", form=form, version=version)


def get_annotated_text(text, annotations):
    """Return HTML-annotated text with entity annotations as tooltips."""
    if not annotations:
        return markupsafe.escape(text)

    # Sort annotations by start index
    annotations = sorted(annotations, key=lambda ann: ann.start)
    annotated_parts = []
    last_index = 0
    for ann in annotations:
        # Add text before the annotation
        annotated_parts.append(markupsafe.escape(text[last_index:ann.start]))

        # Create the annotated span
        match = ann.matches[0]
        match_curie = match.term.get_curie()
        additional_groundings = ", ".join(
            curie for curie in match.get_grounding_dict().keys()
            if curie != match_curie
        )
        title_attr = (f"Grounding: {match_curie} ({match.term.entry_name}), "
                      f"Score: {match.score:.4f}")
        if additional_groundings:
            title_attr += f", Additional groundings: {additional_groundings}"
        span = (f'<span class="annotated-entity" '
                f'title="{markupsafe.escape(title_attr)}">'
                f'{markupsafe.escape(ann.text)}</span>')
        annotated_parts.append(span)
        last_index = ann.end

    # Add remaining text after the last annotation
    annotated_parts.append(markupsafe.escape(text[last_index:]))
    return markupsafe.Markup("".join(annotated_parts))



import os
import boto3
import pystow
import logging
import botocore
from gilda import __version__

logger = logging.getLogger(__name__)

HERE = os.path.abspath(os.path.dirname(__file__))
MESH_MAPPINGS_PATH = os.path.join(HERE, 'mesh_mappings.tsv')

resource_dir = pystow.join('gilda', __version__)

GROUNDING_TERMS_BASE_NAME = 'grounding_terms.tsv.gz'
GROUNDING_TERMS_PATH = os.path.join(resource_dir, GROUNDING_TERMS_BASE_NAME)


# Popular organisms per UniProt, see
# https://www.uniprot.org/help/filter_options
popular_organisms = ['9606', '10090', '10116', '9913', '7955', '7227',
                     '6239', '44689', '3702', '39947', '83333', '224308',
                     '559292']

organism_labels = {
    '9606': 'Homo sapiens',
    '10090': 'Mus musculus',
    '10116': 'Rattus norvegicus',
    '9913': 'Bos taurus',
    '7955': 'Danio rerio',
    '7227': 'Drosophila melanogaster',
    '6239': 'Caenorhabditis elegans',
    '44689': 'Dictyostelium discoideum',
    '3702': 'Arabidopsis thaliana',
    '39947': 'Oryza sativa',
    '83333': 'Escherichia coli',
    '224308': 'Bacillus subtilis',
    '559292': 'Saccharomyces cerevisiae',
}

# NOTE: these are not all exact mappings..
# Several mappings here are to the closest match which works correctly
# in this setting but isn't generally speaking a valid xref.
taxonomy_to_mesh = {
    '9606': 'D006801',
    '10090': 'D051379',
    '10116': 'D051381',
    '9913': 'D002417',
    '7955': 'D015027',
    '7227': 'D004331',
    '6239': 'D017173',
    '44689': 'D004023',
    '3702': 'D017360',
    '39947': 'D012275',
    '83333': 'D048168',
    '224308': 'D001412',
    '559292': 'D012441',
}
mesh_to_taxonomy = {v: k for k, v in taxonomy_to_mesh.items()}


def _download_from_s3(path, base_name):
    config = botocore.client.Config(signature_version=botocore.UNSIGNED)
    s3 = boto3.client('s3', config=config)
    tc = boto3.s3.transfer.TransferConfig(use_threads=False)
    full_key = '%s/%s' % (__version__, base_name)
    out_file = os.path.join(path, base_name)
    s3.download_file('gilda', full_key, out_file, Config=tc)
    return out_file


def get_grounding_terms():
    base_name = GROUNDING_TERMS_BASE_NAME
    full_path = GROUNDING_TERMS_PATH
    if not os.path.exists(full_path):
        logger.info('Downloading grounding terms from S3.')
        out_file = _download_from_s3(resource_dir, base_name)
        logger.info('Saved grounding terms into: %s' % out_file)
    return full_path


def get_gilda_models():
    base_name = 'gilda_models.json.gz'
    full_path = os.path.join(resource_dir, base_name)
    if not os.path.exists(full_path):
        logger.info('Downloading disambiguation models from S3.')
        out_file = _download_from_s3(resource_dir, base_name)
        logger.info('Saved disambiguation models into: %s' % out_file)
    return full_path



from werkzeug.local import LocalProxy

from flask import current_app

__all__ = [
    "grounder",
]

# The way that local proxies work is that when the app gets
# instantiated, you can stick objects into the `app.config`
# dictionary, then the local proxy lets you access them through
# a fake "current_app" object.
grounder = LocalProxy(lambda: current_app.config["grounder"])



import os
import json
import time
import pickle
import argparse
import functools
from multiprocessing import Pool
from collections import Counter
from adeft.modeling.classify import AdeftClassifier
from indra.util import batch_iter
from indra_db.util import get_db
from indra_db.util.content_scripts import get_text_content_from_text_refs
from indra.literature import pubmed_client
from indra.literature.adeft_tools import universal_extract_text
from find_ambiguities import *

db = get_db('primary')


def get_text_content(pmid):
    content = get_text_content_from_text_refs(text_refs={'PMID': pmid},
                                              db=db)
    if content:
        text = universal_extract_text(content)
        return text
    return None


def get_pmids(ambig_terms):
    term_pmids = {}
    pmid_counter = Counter()
    for term in ambig_terms:
        key = (term.db, term.id)
        if term.db == 'HGNC':
            gene = term.entry_name
            try:
                term_pmids[key] = pubmed_client.get_ids_for_gene(gene)
            except ValueError:
                print('Could not get PMIDs for gene: %s' % gene)
                term_pmids[key] = []
            pmid_counter.update(term_pmids[key])
            time.sleep(0.5)
        elif term.db == 'MESH':
            pmids = pubmed_client.get_ids_for_mesh(term.id, major_topic=False)
            if len(pmids) > 1000:
                pmids = pubmed_client.get_ids_for_mesh(term.id,
                                                       major_topic=True)
            term_pmids[key] = pmids[:1000]
            pmid_counter.update(term_pmids[key])
            time.sleep(0.5)
        else:
            print('Unhandled ambiguous term: %s' % str(key))
    term_pmids = {k: [p for p in pmids if pmid_counter[p] == 1]
                  for k, pmids in term_pmids.items()}
    return term_pmids


def get_all_pmids(ambigs):
    all_pmids = [(a, get_pmids(a)) for a in ambigs]
    return all_pmids


def get_texts_for_term(key, pmids):
    texts = []
    print('Loading %d PMIDs for %s' % (len(pmids), str(key)))
    for pmid in pmids:
        txt = get_text_content(pmid)
        if txt:
            texts.append(txt)
    print('Loaded %d texts for %s' % (len(texts), str(key)))
    if texts and len(texts) <= 5:
        print('Splitting texts for %s' % str(key))
        texts = split_texts(texts, 6)
        print('Now have %s texts for %s' % (len(texts), str(key)))
    return texts


def get_papers(ambig_terms, term_pmids=None):
    if not term_pmids:
        term_pmids = get_pmids(ambig_terms)
    texts = []
    labels = []
    for key, pmids in term_pmids.items():
        texts1 = get_texts_for_term(key, pmids)
        texts += texts1
        labels += ['%s:%s' % key for _ in texts1]
    return texts, labels


def split_texts(texts, nmin):
    while len(texts) < nmin:
        texts = sorted(texts, key=lambda x: len(x), reverse=True)
        texts = [texts[0][:int(len(texts[0])/2)],
                 texts[0][int(len(texts[0])/2):]] + texts[1:]
    return texts


def rank_ambiguities(ambigs, str_counts):
    sorted_ambigs = sorted(ambigs, key=lambda x: str_counts.get(x[0].text, 0),
                           reverse=True)
    return sorted_ambigs


def learn_model(ambig_terms_pmids, params):
    ambig_terms, term_pmids = ambig_terms_pmids

    print()
    terms_str = '\n> ' + '\n> '.join(str(t) for t in ambig_terms)
    print('Learning model for: %s\n=======' % terms_str)
    texts, labels = get_papers(ambig_terms, term_pmids)
    if len(set(labels)) < 2:
        print('Could not get enough labels for more than one class, skipping.')
        return None

    label_counts = Counter(labels)
    if any([v <= 5 for v in label_counts.values()]):
        print('Could not get enough labels for at least one entry, skipping.')
        return None
    cl = AdeftClassifier([ambig_terms[0].text], list(set(labels)))
    cl.cv(texts, labels, params, cv=5)
    print(cl.stats)
    cl_model_info = cl.get_model_info()
    return {'cl': cl_model_info, 'ambig': ambig_terms}


def learn_batch(ambig_terms_batch):
    models = {}
    for ambig_terms in ambig_terms_batch:
        entity_text = ambig_terms[0].text
        res = learn_model(ambig_terms, param_grid)
        models[entity_text] = res
    return models


def get_filter_ambigs(str_counts):
    ambigs = get_ambiguities()
    ambigs = filter_out_duplicates(ambigs)
    ambigs = filter_out_shared_prefix(ambigs)
    ambigs = filter_out_mesh_proteins(ambigs)
    ambigs = rank_ambiguities(ambigs, str_counts)
    return ambigs


with open('raw_agent_text_count.json') as fh:
    str_counts = json.load(fh)


if __name__ == '__main__':
    # Initialization
    parser = argparse.ArgumentParser()
    parser.add_argument('--nproc', default=1, type=int)
    args = parser.parse_args()
    pmid_cache_file = 'all_pmids.pkl'
    if os.path.exists(pmid_cache_file):
        # Cached PMIDs
        with open('all_pmids.pkl', 'rb') as fh:
            all_ambigs_pmids = pickle.load(fh)
    else:
        ambigs = get_filter_ambigs(str_counts)
        all_ambigs_pmids = get_all_pmids(ambigs)

    # Construct list of ambiguities
    #print('Found a total of %d ambiguities.' % len(ambigs))

    # Learn models
    param_grid = {'C': [10.0], 'max_features': [100, 1000],
                  'ngram_range': [(1, 2)]}
    models = {}
    if args.nproc == 1:
        for idx, ambig_terms_batch in \
                enumerate(batch_iter(all_ambigs_pmids, 10)):
            pickle_name = 'gilda_ambiguities_hgnc_mesh_%d.pkl' % idx
            models = learn_batch(ambig_terms_batch)
            with open(pickle_name, 'wb') as fh:
                pickle.dump(models, fh)
    else:
        pool = Pool(args.nproc)
        fun = functools.partial(learn_model, params=param_grid)
        pkl_idx = 0
        models = {}
        for count, model in enumerate(pool.imap_unordered(fun,
                                                          all_ambigs_pmids,
                                                          chunksize=10)):
            print('#### %d ####' % count)
            if model is None:
                print('Model is None, skipping')
            else:
                models[model['ambig'][0].text] = model
            if (count + 1) % 100 == 0:
                pickle_name = 'gilda_ambiguities_hgnc_mesh_%d.pkl' % pkl_idx
                with open(pickle_name, 'wb') as fh:
                    pickle.dump(models, fh)
                pkl_idx += 1
                models = {}
        pool.close()
        pool.join()




"""
Utilities for greek characters from https://gist.github.com/beniwohli/765262.
"""

__all__ = [
    "greek_alphabet",
    "greek_to_latin",
]

#: A mapping from unicode greek characters to latin names for greek characters
greek_alphabet = {
    u'\u0391': 'Alpha',
    u'\u0392': 'Beta',
    u'\u0393': 'Gamma',
    u'\u0394': 'Delta',
    u'\u0395': 'Epsilon',
    u'\u0396': 'Zeta',
    u'\u0397': 'Eta',
    u'\u0398': 'Theta',
    u'\u0399': 'Iota',
    u'\u039A': 'Kappa',
    u'\u039B': 'Lamda',
    u'\u039C': 'Mu',
    u'\u039D': 'Nu',
    u'\u039E': 'Xi',
    u'\u039F': 'Omicron',
    u'\u03A0': 'Pi',
    u'\u03A1': 'Rho',
    u'\u03A3': 'Sigma',
    u'\u03A4': 'Tau',
    u'\u03A5': 'Upsilon',
    u'\u03A6': 'Phi',
    u'\u03A7': 'Chi',
    u'\u03A8': 'Psi',
    u'\u03A9': 'Omega',
    u'\u03B1': 'alpha',
    u'\u03B2': 'beta',
    u'\u03B3': 'gamma',
    u'\u03B4': 'delta',
    u'\u03B5': 'epsilon',
    u'\u03B6': 'zeta',
    u'\u03B7': 'eta',
    u'\u03B8': 'theta',
    u'\u03B9': 'iota',
    u'\u03BA': 'kappa',
    u'\u03BB': 'lamda',
    u'\u03BC': 'mu',
    u'\u03BD': 'nu',
    u'\u03BE': 'xi',
    u'\u03BF': 'omicron',
    u'\u03C0': 'pi',
    u'\u03C1': 'rho',
    u'\u03C3': 'sigma',
    u'\u03C4': 'tau',
    u'\u03C5': 'upsilon',
    u'\u03C6': 'phi',
    u'\u03C7': 'chi',
    u'\u03C8': 'psi',
    u'\u03C9': 'omega',
}


#: A mapping from latin names for greek characters to latin characters
greek_to_latin = {
   'alpha': 'a',
   'Alpha': 'A',
   'beta': 'b',
   'Beta': 'B',
   'gamma': 'c',
   'Gamma': 'C',
   'delta': 'd',
   'Delta': 'D',
}



import argparse
import logging
from textwrap import dedent
from typing import Optional

from flask import Flask, abort, jsonify, redirect, request
from flask_restx import Api, Resource, fields

from gilda import __version__ as version
from gilda.grounder import GrounderInput, Grounder
from gilda.app.proxies import grounder

# NOTE: the Flask REST-X API has to be declared here, below the home endpoint
# otherwise it reserves the / base path.

logger = logging.getLogger(__name__)

api = Api(title="Gilda",
          description="A service for grounding entity strings",
          version=version,
          license="Code available under the BSD 2-Clause License",
          contact="Gyori Lab",
          contact_email="b.gyori@northeastern.edu",
          contact_url="https://gyorilab.github.io",
          doc='/apidocs',
          )

base_ns = api.namespace('Gilda API', 'Gilda API', path='/')

grounding_input_model = api.model(
    "GroundingInput",
    {'text': fields.String(example='EGF-receptor',
                           required=True,
                           description='The entity string to be grounded.'),
     'context': fields.String(example='The EGFR-receptor binds EGF.',
                              required=False,
                              description='Surrounding text as context to aid'
                                          ' disambiguation when applicable.'),
     'organisms': fields.List(fields.String, example=['9606'],
                              description='An optional list of taxonomy '
                                          'species IDs defining a priority list'
                                          ' in case an entity string can be '
                                          'resolved to multiple'
                                          'species-specific genes/proteins.',
                              required=False)}
)

term_model = api.model(
    "Term",
    {'norm_text': fields.String(
        description='The normalized text corresponding to the text entry, '
                    'used for lookups.',
        example='egf receptor'),
     'text': fields.String(
         description='The text entry that was matched.',
         example='EGF receptor'
     ),
     'db': fields.String(
         description='The database / namespace corresponding to the '
                     'grounded term.',
         example='HGNC'
     ),
     'id': fields.String(
         description='The identifier of the grounded term within the '
                     'database / namespace.',
         example='3236'
     ),
     'entry_name': fields.String(
         description='The standardized name corresponding to the grounded '
                     'term.',
         example='EGFR'
     ),
     'status': fields.String(
         description='The relationship of the text entry to the grounded '
                     'term, e.g., synonym.',
         example='curated'
     ),
     'source': fields.String(
         description='The source from which the term was obtained.',
         example='famplex'
     ),
     'organism': fields.String(
         description='If the term is a gene/protein, this field provides '
                     'the taxonomy identifier of the species to which '
                     'it belongs.',
         example='9606'
     ),
     'source_db': fields.String(
         description='In some cases the term\'s db/id was mapped from another '
                     'db/id pair given in the original source. If this is the '
                     'case, this field provides the original source db.'),
     'source_id': fields.String(
         description='In some cases the term\'s db/id was mapped from another '
                     'db/id pair given in the original source. If this is the '
                     'case, this field provides the original source ID.')
    }
)

scored_match_model = api.model(
    "ScoredMatch",
    {'term': fields.Nested(term_model, description='The term that was matched'),
     'url': fields.String(
         description='bioregistry.io URL for the matched term.',
         example='https://bioregistry.io/hgnc:3236'
     ),
     'score': fields.Float(
         description='The score assigned to the matched term.',
         example=0.9845
     ),
     'match': fields.Nested(api.model('Match', {}),
         description='Additional metadata about the nature of the match.'
     ),
     'subsumed_terms': fields.List(fields.Nested(term_model),
         description='In some cases multiple terms with the same db/id '
                     'matched the input string, potentially with different '
                     'scores, and only the first one is exposed in the '
                     'scored match\'s term attribute (see above). This field '
                     'provides additional terms with the same db/id that '
                     'matched the input for additional traceability.')
     }
)

get_names_input_model = api.model(
    "GetNamesInput",
    {'db': fields.String(
         description="Capitalized name of the database for the grounding, "
                     "e.g. HGNC.",
         required=True,
         example='HGNC'),
     'id': fields.String(
         description="Identifier within the given database",
         required=True,
         example='3236'
     ),
     'status': fields.String(
         description="If provided, only entity texts of the given status are "
                     "returned (e.g., curated, name, synonym, former_name).",
         required=False,
         enum=['curated', 'name', 'synonym', 'former_name'],
         example='synonym'
     ),
     'source': fields.String(
         description="If provided, only entity texts collected from the given "
                     "source are returned.This is useful if terms grounded to "
                     "IDs in a given database are collected from multiple "
                     "different sources.",
         required=False,
         example='uniprot'
    )
    }
)

ner_result_model = api.model('NERResult', {
    'text': fields.String(description='Matched text'),
    'start': fields.Integer(description='Start index of the match'),
    'end': fields.Integer(description='End index of the match'),
    'matches': fields.List(fields.Nested(scored_match_model))
})

ner_input_model = api.model('NERInput', {
    'text': fields.String(required=True, description='Text on which to perform'
                                                     ' NER',
                          example='The EGF receptor binds EGF which is an '
                                  'interaction important in cancer.'),
    'organisms': fields.List(fields.String, example=['9606'],
                             description='An optional list of taxonomy '
                                         'species IDs defining a priority list'
                                         ' in case an entity string can be '
                                         'resolved to multiple'
                                         'species-specific genes/proteins.',
                             required=False),
    'namespaces': fields.List(fields.String,
                              description='A list of namespaces to pass to '
                                          'the grounder to restrict the '
                                          'matches to. By default, '
                                          'no restriction is applied',
                              example=['HGNC', 'MESH'],
                              required=False),
    'context_text': fields.String(required=False, description='Additional '
                                                              'context for '
                                                              'disambiguation',
                                  example='The EGF receptor binds EGF which is an interaction'
                                          'important in cancer.'),
})

names_model = fields.List(
    fields.String,
    example=['EGF receptor', 'EGFR', 'ERBB1', 'Proto-oncogene c-ErbB-1'])

models_model = fields.List(
    fields.String,
    example=['A4', 'ABC1', 'p180'])


@base_ns.route('/ground', methods=['POST'])
class Ground(Resource):
    @base_ns.response(200, "Grounding results", [scored_match_model])
    @base_ns.expect(grounding_input_model)
    def post(self):
        """Perform grounding on a given entity text.

        Returns a list of scored grounding matches for the given entity text.
        The returned value is a list with each entry being a scored match.
        Each scored match contains a term which was matched, and each term
        contains a db and id constituting a grounding. An empty list
        return value means that no grounding matches were found for the input.
        """
        if request.json is None:
            abort(415, 'Missing application/json header.')
        # Get input parameters
        text = request.json.get('text')
        context = request.json.get('context')
        organisms = request.json.get('organisms')
        scored_matches = grounder.ground(text, context=context,
                                         organisms=organisms)
        res = [sm.to_json() for sm in scored_matches]
        return jsonify(res)


@base_ns.route('/ground_multi', methods=['POST'])
class GroundMulti(Resource):
    @base_ns.response(200, "Grounding results", [[scored_match_model]])
    @base_ns.expect([grounding_input_model])
    def post(self):
        """Return a list of grounding matches for a list of inputs.

        This endpoint is useful for batch processing of inputs. The input
        is a list with each entry containing a text key as well as optional
        context and organism keys. The returned value is a list corresponding
        to each input in order, each entry of the returned list being a list
        of scored matches.
        """
        if request.json is None:
            abort(415, 'Missing application/json header.')
        # Get input parameters
        all_matches = []
        for input in request.json:
            text = input.get('text')
            context = input.get('context')
            organisms = input.get('organisms')
            scored_matches = grounder.ground(text, context=context,
                                             organisms=organisms)
            all_matches.append([sm.to_json() for sm in scored_matches])
        return jsonify(all_matches)


@base_ns.route('/names', methods=['POST'])
@base_ns.route('/get_names', methods=['POST'])
class GetNames(Resource):
    @base_ns.response(200, "Get names result", names_model)
    @base_ns.expect(get_names_input_model)
    def post(self):
        """Return all known entity texts for a grounding.

        This endpoint can be used as a reverse lookup to find out what entity
        texts (names, synonyms, etc.) are known for a given grounded entity.
        """
        if request.json is None:
            abort(415, 'Missing application/json header.')
        # Get input parameters
        kwargs = {key: request.json.get(key) for key in {'db', 'id', 'status',
                                                         'source'}}
        names = grounder.get_names(**kwargs)
        return jsonify(names)


@base_ns.route('/models', methods=['GET', 'POST'])
class GetModels(Resource):
    @base_ns.response(200, "Get models result", models_model)
    def post(self):
        """Return a list of texts with Gilda disambiguation models.

        Gilda makes available more than one thousand disambiguation models
        between synonyms shared by multiple genes. This endpoint returns
        the list of entity texts for which such a model is available.
        """
        return jsonify(grounder.get_models())

    def get(self):
        """Return a list of texts with Gilda disambiguation models.

        Gilda makes available more than one thousand disambiguation models
        between synonyms shared by multiple genes. This endpoint returns
        the list of entity texts for which such a model is available.
        """
        return jsonify(grounder.get_models())


@base_ns.route('/annotate', methods=['POST'])
class Annotate(Resource):
    @base_ns.response(200, "NER results", [ner_result_model])
    @base_ns.expect(ner_input_model)
    def post(self):
        """Perform named entity recognition on the given text.

        This endpoint can be used to perform named entity recognition (NER)
        using Gilda.
        """
        from gilda.ner import annotate

        if request.json is None:
            abort(415, 'Missing application/json header.')

        text = request.json.get('text')
        context_text = request.json.get('context_text')
        organisms = request.json.get('organisms')
        namespaces = request.json.get('namespaces')

        results = annotate(text, organisms=organisms if organisms else None,
                           namespaces=namespaces if namespaces else None,
                           context_text=context_text)
        return jsonify([annotation.to_json() for annotation in results])


def get_app(terms: Optional[GrounderInput] = None, *, ui: bool = True) -> Flask:
    app = Flask(__name__)
    app.config['RESTX_MASK_SWAGGER'] = False
    app.config['WTF_CSRF_ENABLED'] = False
    app.config['SWAGGER_UI_DOC_EXPANSION'] = 'list'
    app.config["grounder"] = Grounder(terms=terms)

    if not ui:
        _mount_home_redirect(app)
    else:
        try:
            import importlib.metadata as importlib_metadata
        except ImportError:
            import importlib_metadata

        try:
            from flask_bootstrap import Bootstrap

            bootstrap_version = importlib_metadata.version("flask_bootstrap")
            if "3.3.7.1" != bootstrap_version:
                raise ImportError(
                    dedent(
                        """\
                    The wrong flask-bootstrap is installed, therefore the UI
                    can not be enabled. Please run the following commands in
                    the shell:

                      pip uninstall flask-bootstrap bootstrap-flask
                      pip install flask-bootstrap
                    """
                    )
                )

            from gilda.app.ui import ui_blueprint
        except ImportError as e:
            logger.error('Could not import UI blueprint: %s', e)
            _mount_home_redirect(app)
        else:
            Bootstrap(app)
            app.register_blueprint(ui_blueprint, url_prefix="/")

    # has to be put after defining the UI blueprint otherwise it reserves "/"
    api.init_app(app)
    return app


def _mount_home_redirect(app):
    @app.route("/")
    def home_redirect():
        """Redirect the home url to the API documentation."""
        return redirect("/apidocs")


def parse_args():
    parser = argparse.ArgumentParser(
        description='Run the grounding app.')
    parser.add_argument('--host', default='0.0.0.0')
    parser.add_argument('--port', default=8001, type=int)
    parser.add_argument('--terms')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    app = get_app(args.terms)
    app.run(args.host, args.port, threaded=False)



from gilda.term import Term, get_bioregistry_url, get_identifiers_url


def test_standalone_get_url():
    assert get_bioregistry_url('UP', 'P12345') == \
        'https://bioregistry.io/uniprot:P12345'
    assert get_bioregistry_url('HGNC', '12345') == \
        'https://bioregistry.io/hgnc:12345'
    assert get_identifiers_url('CHEBI', 'CHEBI:12345') == \
        'https://identifiers.org/CHEBI:12345'
    assert get_bioregistry_url('CHEBI', 'CHEBI:12345') == \
           'https://bioregistry.io/chebi:12345'


def test_term_get_url():
    term = Term(db='CHEBI', id='CHEBI:12345', entry_name='X',
                norm_text='x', text='X', source='test', status='name')
    assert term.get_curie() == 'chebi:12345'
    assert term.get_curie(style='identifiers') == 'CHEBI:12345'
    assert term.get_identifiers_url() == \
        'https://identifiers.org/CHEBI:12345'
    assert term.get_bioregistry_url() == \
        'https://bioregistry.io/chebi:12345'
    assert term.get_groundings() == {(term.db, term.id)}
    assert term.get_namespaces() == {term.db}


def test_term_source_db_id():
    term = Term('mitochondria', 'Mitochondria', 'GO', 'GO:0005739',
                'mitochondrion', 'synonym', 'mesh', None, 'MESH', 'D008928')
    assert term.source_db == 'MESH'
    assert term.source_id == 'D008928'
    assert term.get_groundings() == {(term.db, term.id),
                                     (term.source_db, term.source_id)}

    assert term.get_namespaces() == {term.db, term.source_db}



"""This module implements an optional SQLite database back-end for Gilda
as an alternative to loading Terms directly into memory."""

import os
import json
import sys
import logging
import sqlite3
from gilda.term import Term
from . import resource_dir

logger = logging.getLogger('gilda.resources.sqlite_adapter')


class SqliteEntries:
    """A class exposing lists of Terms similar to a string-keyed dict.

    From the perspective of a Grounder instance, instances of this class
    have an interface similar to a dict ot lists of Terms and can therefore
    be used seamlessly as a Grounder instance's entries attribute.

    Parameters
    ----------
    db : str
        A path to a SQLite database file.
    """
    def __init__(self, db):
        self.db = db
        self.conn = None

    def get_connection(self):
        if self.conn:
            return self.conn
        self.conn = sqlite3.connect(self.db)
        return self.conn

    def get(self, key, default=None):
        res = self.get_connection().execute(
            "SELECT terms FROM terms WHERE norm_text=?", (key,))
        result = res.fetchone()
        if not result:
            return default
        return [Term(**j) for j in json.loads(result[0])]

    def values(self):
        res = self.get_connection().execute("SELECT terms FROM terms")
        for result in res.fetchall():
            yield [Term(**j) for j in json.loads(result[0])]

    def __getitem__(self, item):
        res = self.get(item)
        if res is None:
            raise KeyError(item)
        return res

    def __len__(self):
        res = self.get_connection().execute("SELECT COUNT(norm_text) FROM terms")
        return res.fetchone()[0]

    def __iter__(self):
        res = self.get_connection().execute("SELECT norm_text FROM terms")
        for norm_text, in res.fetchall():
            yield norm_text


def build(grounding_entries, path=None):
    """Build a SQLite database file from a set of grounding entries.

    Parameters
    ----------
    grounding_entries : dict[str, list[Term]]
        A grounding entries data structure from which the DB is generated.
    path : Optional[str, Path]
        Optional path to the output file which should use the .db extension.
        If not given, the .db file is generated in Gilda's default resources
        folder.
    """
    path = path if path else os.path.join(resource_dir, 'grounding_terms.db')
    logger.info('Starting SQLite database at %s' % path)
    conn = sqlite3.connect(path)
    cur = conn.cursor()

    # Create the table
    logger.info('Creating the table')
    q = "CREATE TABLE terms (norm_text text not null primary key, terms text)"
    cur.execute(q)

    # Insert terms
    logger.info('Inserting terms')
    q = "INSERT INTO terms (norm_text, terms) VALUES (?, ?)"
    for norm_text, terms in grounding_entries.items():
        cur.execute(q, (norm_text, json.dumps([t.to_json() for t in terms])))

    # Build index
    logger.info('Making index')
    q = "CREATE INDEX norm_index ON terms (norm_text);"
    cur.execute(q)
    conn.commit()
    conn.close()


if __name__ == '__main__':
    from gilda.grounder import Grounder

    path = sys.argv[1] if len(sys.argv) > 1 else None
    logger.info('Loading default grounder')
    gr = Grounder()
    build(gr.entries, path=path)



from gilda.process import depluralize, replace_greek_spelled_out, \
    replace_roman_arabic, replace_greek_uni, replace_greek_latin, normalize


def test_depluralize():
    assert depluralize('BRAF') == [('BRAF', 'non_plural')]
    assert depluralize('apoptosis') == [('apoptosis', 'non_plural')]
    assert depluralize('mosquitoes') == [('mosquito', 'plural_oes'),
                                         ('mosquitoe', 'plural_s')]
    assert depluralize('antibodies') == [('antibody', 'plural_ies'),
                                         ('antibodie', 'plural_s')]
    assert depluralize('branches') == [('branch', 'plural_es'),
                                       ('branche', 'plural_s')]
    assert depluralize('CDs') == [('CD', 'plural_caps_s')]
    assert depluralize('receptors') == [('receptor', 'plural_s')]
    assert depluralize('kinases') == [('kinas', 'plural_es'),
                                      ('kinase', 'plural_s')]


def test_greek():
    assert replace_greek_spelled_out('interferon-γ') == 'interferon-gamma'
    assert replace_greek_uni('interferon-gamma') == 'interferon-γ'
    assert replace_greek_latin('interferon-beta') == 'interferon-b'


def test_roman_arabic():
    assert replace_roman_arabic('xx-1') == 'xx-I'
    assert replace_roman_arabic('xx-10') == 'xx-X'
    assert replace_roman_arabic('x1x') == 'x1x'
    assert replace_roman_arabic('xx-I') == 'xx-1'
    assert replace_roman_arabic('xx viii') == 'xx 8'
    assert replace_roman_arabic('xx-iX') == 'xx-9'


def test_normalize():
    assert normalize('Löfgren’s syndrome') == 'lofgren\'s syndrome'
    assert normalize('βAR') == 'βar'